#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2009-2012 Valentin Haenel <valentin.haenel@gmx.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

""" Python interface to the CRS 'OptiCAL' photometer.

    Overview
    ========

    The 'OptiCAL' is a photometer that is produced by Cambridge Research
    Systems (CRS). This device is a standard tool for gamma-calibration of
    display devices in vision research. This package provides a free-software
    replacement for the Windows-software distributed by the manufacturer that
    allows querying an OptiCAL via a serial connection. `pyoptical` can be used
    as a library in third-party applications or as a standalone command line
    tool.

    This module provides the `OptiCAL` class and some supporting code. The
    command line wrapper for this module can be found in the 'pyoptical'
    script.

    Examples
    --------

    Command line:
        ``pyoptical --number=15 --interval=750 /dev/ttyUSB0``

        Make 15 measurements with an interval of 750 ms between them.

        usage:
            ``pyoptical [-i interval] [-n number ] [-r] com-port``

        For more information try, ``pyoptical --help``. For more information
        about the ``com-port`` argument see: `Notes about the com-port`_.

    Library:
        >>> import pyoptical
        >>> op = pyoptical.OptiCAL('dev/dev/ttyUSB0')
        >>> try:
        >>>     op.read_luminance()
        >>> except pyoptical.NACKException as e:
        >>>     print(e)

    Notes about the com-port
    ------------------------

    The com_port argument for the constructor may vary depending on both
    your operating system and how you connect to the OptiCAL. This code
    was developed using a usb-to-serial adapter that contains a PL2303
    chipset manufactured by Prolific:
    http://www.prolific.com.tw/eng/Products.asp?ID=59. The following
    sections outline how to access the OptiCAL using pyoptical and a
    usb-to-serial adapter containing the prolific chipset. We have not
    tried this code using a raw serial port, but would be very
    interested to hear from you if you do.

    Linux (Ubuntu Hardy / Debian Lenny):
        Support for the PL2303 chipset is compiled into the kernel, and
        the device is automatically recognised. You could check 'dmesg'
        for the following output::

            usb 2-1: new full speed USB device using uhci_hcd and address 4
            usb 2-1: configuration #1 chosen from 1 choice
            pl2303 2-1:1.0: pl2303 converter detected
            usb 2-1: pl2303 converter now attached to ttyUSB0

        In this case the com_port string is simply ``/dev/ttyUSB0``

    Mac OSX (10.5.8 Leopard)
        Support for the PL2303 chipset is provided by the following
        open source driver: http://osx-pl2303.sourceforge.net/

        In this case the com-port string would be something along the
        lines of ``/dev/tty.PL2303-xxx``, for example:
        ``/dev/tty.PL2303-000031FD``

    Windows (XP)
        The manufacturer of your usb-to-serial adapter should provide
        you with drivers.

        In this case the com-port string would be something like:
        ``COM2``, check the device manager for the number of the COM port.

    Other Operating Systems and Adapters:
        This code has two limitations, most importantly pyserial must
        support your platform. Secondly, if you wish to use a
        usb-to-serial adapter a driver for your target operating system
        must be available from the manufacturer or possibly a third
        party (for example and open source driver).

    Notes about possible exceptions
    -------------------------------

    There are three types of exceptions that can happen:

    - `OptiCALException`
    - `NACKException`
    - `TimeoutException`

    The `OptiCALException` is the base class for all exceptions in this
    module, and it is used as a general purpose exception to signify
    errors on the part of the programmer, do not quietly except these.

    The `NACKException` is raised when the OptiCAL responds with a ``NACK``
    byte. It does this either if the command was not understood or if
    the command failed. If this happens during initialization, you may
    have to re-initialise the device. If this happens during readout it
    should be safe to try again instead of terminating the program.

    The `TimeoutException` is raised when no answer is received within the
    default timeout length. This might be caused by a number of issues,
    but essentially means that somehow the communication with the
    OptiCAL might be interrupted, for example because it is no longer
    connected to the computer.

    Implementation details
    ----------------------

    The interface is implemented according to the protocol specification in the
    OptiCAL-User-Guide Version 4, 1995 including the following amendments:

    1. To read out the ADC value, an 'L' must be sent instead of an 'R'
    2. The equations to convert from ADC to meaningful units had changed. See
       the source code of ``read_luminance()`` for details.

    The full errata is available from the CSR website:

    http://support.crsltd.com/FileManagement/Download/9f5f62bcb3e64eb8934fe72afb937cb6

    The corrected versions of the conversion formulas can also be found
    in the OptiCAL.py python interface available from the CRS website,
    written by Walter F. Bischof in 2007. However note that this code may
    contain a bug, see
    http://groups.google.com/group/psychopy-users/msg/9cb35590daf3a29b
    for detail.

    The constructor will first perform the initial calibration of the
    device as required by the protocol specification. Next it will read
    out all parameters from the eeprom and store them as private
    variables. And lastly it will put the device into the default mode.

    The initial version of the OptiCAL hardware supported two readout
    modes 'current' and 'voltage'. The device could be used to read
    luminance when in 'current' mode and 'voltage' when in voltage mode.
    Over the years there have been two revisions of the OptiCAL
    hardware, both no longer supported usage as a voltmeter, and thus
    the 'voltage' mode has become redundant. Since version 0.2 this
    interface no longer supports the 'voltage' mode, and the device will
    be put into 'current' mode at startup.

"""

__version__ = "0.4"
__author__ = "Valentin Haenel <valentin.haenel@gmx.de>"
__docformat__ = "restructuredtext en"

import serial


class OptiCAL:
    """ Object to access the OptiCAL """

    _ACK = '\x06'
    _NACK = '\x15'

    def __init__(self, com_port, timeout=5):
        """ initialise OptiCAL

            :Parameters:
                com_port : string
                    name of the com-port
                timeout : float
                    time in seconds to wait for a response

            For more information about the ``com_port`` argument see:
            the '``Notes about the com-port``' section in the module docstring.
        """
        self._phot = serial.Serial(com_port, timeout=timeout)
        self._calibrate()
        self._read_ref_defs()
        self._read_other_defs()
        self._set_current_mode()

    def __str__(self):
        return "Optical found at : " + self._phot.port + "\n" + \
               "Product Type :     " + str(self._product_type) + "\n" + \
               "Optical S/N  :     " + str(self._optical_serial_number) + "\n"\
               "Firmware version : " + str(self._firmware_version) + "\n" \
               "V_ref:             " + str(self._V_ref) + "\n" + \
               "Z_count:           " + str(self._Z_count) + "\n" + \
               "R_feed:            " + str(self._R_feed) + "\n" + \
               "R_gain:            " + str(self._R_gain) + "\n" + \
               "Probe S/N          " + str(self._probe_serial_number) + "\n" +\
               "K_cal:             " + str(self._K_cal) + "\n"

    def _send_command(self, command, description):
        """ send a single command character and read a single response (ACK/NACK)

            :Parameters:
                command : chr
                    the command character
                description : string
                    a string describing the command

        """
        self._phot.write(command)
        ret = self._phot.read()
        _check_return(ret, description)

    def _calibrate(self):
        """ perform initial calibration

            As stated in the OptiCAL user guide, this must be done after
            powering up the device, before any readouts are performed.

        """
        self._send_command('C', "calibrate")

    def _set_current_mode(self):
        """ put the device into 'current' mode """
        self._send_command('I', "set current mode")

    def _read_eeprom_single(self, address):
        """ read contents of eeprom at single address

            note: the ACK byte is truncated

            :Parameters:
                address : int
                    address in the range 0<i<100

            :Return:
                (string) - a byte in the range 0<i<256

        """
        self._phot.write(chr(128 + address))
        ret = self._phot.read(2)
        _check_return(ret, "reading eeprom at address %d" % address)
        # if _check_return does not raise an exception
        return ret[0]

    def _read_eeprom(self, start, stop):
        """ read contents of eeprom between start and stop inclusive

            :Parameters:
                start : int
                    address in the range 0<i<100
                stop : int
                    address in the range 0<i<100

            :Returns:
                (string of bytes) - each character in the range 0<i<255
        """
        return "".join([self._read_eeprom_single(i)
                        for i in range(start, stop + 1)])

    def _read_product_type(self):
        return _to_int(self._read_eeprom(0, 1))

    def _read_optical_serial_number(self):
        return _to_int(self._read_eeprom(2, 5))

    def _read_firmware_version(self):
        return float(_to_int(self._read_eeprom(6, 7))) / 100

    def _read_probe_serial_number(self):
        return int(self._read_eeprom(80, 95))

    def _read_other_defs(self):
        """ read all parameters that do not have a ref definition """
        self._product_type = self._read_product_type()
        self._optical_serial_number = self._read_optical_serial_number()
        self._firmware_version = self._read_firmware_version()
        self._probe_serial_number = self._read_probe_serial_number()

    def _read_V_ref(self):
        """ reference voltage in microV """
        return _to_int(self._read_eeprom(16, 19))

    def _read_Z_count(self):
        """ zero error in ADC counts """
        return _to_int(self._read_eeprom(32, 35))

    def _read_R_feed(self):
        """ feedback resistor in Ohm """
        return _to_int(self._read_eeprom(48, 51))

    def _read_R_gain(self):
        """ voltage gain resistor in Ohm """
        return _to_int(self._read_eeprom(64, 67))

    def _read_K_cal(self):
        """ probe calibration in fA/cd/m**2 """
        return _to_int(self._read_eeprom(96, 99))

    def _read_ref_defs(self):
        """ read all parameters with a ref definition """
        self._V_ref = self._read_V_ref()
        self._Z_count = self._read_Z_count()
        self._R_feed = self._read_R_feed()
        self._R_gain = self._read_R_gain()
        self._K_cal = self._read_K_cal()

    def _read_adc(self):
        """ read and adjust the ADC value """
        self._phot.write('L')
        ret = self._phot.read(4)
        _check_return(ret, "reading adc value")
        # truncate the ACK
        ret = ret[:-1]
        # obtain an integer value from the bytes
        adc = _to_int(ret)
        return adc - self._Z_count - 524288

    def read_luminance(self):
        """ the luminance in cd/m**2 """
        ADC_adjust = self._read_adc()
        numerator = (float(ADC_adjust) / 524288) * self._V_ref * 1.e-6
        denominator = self._R_feed * self._K_cal * 1.e-15
        return max(0.0, numerator / denominator)


def _to_int(byte_string):
    """ convert a string of bytes(in least significant byte order) to int """
    return int(byte_string[::-1].encode('hex'), 16)


def _check_return(ret, description):
    """ check the return value of a read, raise exception if its not OK """
    if ret == "":
        raise TimeoutException(description)
    # check the last byte only
    elif ret[-1] == OptiCAL._NACK:
        raise NACKException(description)


class OptiCALException(Exception):
    """ base exception for all OptiCAL exceptions """


class NACKException(OptiCALException):
    """ raised when the OptiCAL sends a NACK byte to signify an error"""

    def __str__(self):
        return "OptiCAL sent a NACK while trying to: %s" % self.message


class TimeoutException(OptiCALException):
    """ raised when the OptiCAL does not respond within the timeout limit """

    def __str__(self):
        return "OptiCAL timeout while trying to: %s" % self.message



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys

from pkg_resources import parse_version

from .base import DataHandler
from .experiment import ExperimentHandler
from .trial import TrialHandler, TrialHandler2, TrialHandlerExt, TrialType
from .staircase import (StairHandler, QuestHandler, PsiHandler,
                        MultiStairHandler)

if sys.version_info.major == 3 and sys.version_info.minor >= 6:
    from .staircase import QuestPlusHandler

from .utils import (checkValidFilePath, isValidVariableName, importTrialTypes,
                    sliceFromString, indicesFromString, importConditions,
                    createFactorialTrialList, bootStraps, functionFromStaircase,
                    getDateStr)

from .fit import (FitFunction, FitCumNormal, FitLogistic, FitNakaRushton,
                  FitWeibull)

try:
    # import openpyxl
    import openpyxl
    if parse_version(openpyxl.__version__) >= parse_version('2.4.0'):
        # openpyxl moved get_column_letter to utils.cell
        from openpyxl.utils.cell import get_column_letter
    else:
        from openpyxl.cell import get_column_letter
    from openpyxl.reader.excel import load_workbook
    haveOpenpyxl = True
except ImportError:
    haveOpenpyxl = False

try:
    import xlrd
    haveXlrd = True
except ImportError:
    haveXlrd = False



from psychopy import visual, layout
from .test_basevisual import _TestColorMixin, _TestUnitsMixin

class TestTarget(_TestUnitsMixin):
    # Pixel which is the border color
    borderPoint = (0, 55)
    borderUsed = True
    # Pixel which is the fill color
    fillPoint = (0, 30)
    fillUsed = False
    # Pixel which is the fore color
    forePoint = (0, 0)
    foreUsed = False

    @classmethod
    def setup_class(cls):
        cls.win = visual.Window(size=(128, 128))
        cls.obj = visual.TargetStim(cls.win, "TargetStim", units='pix', pos=(-64, 64),
                                    innerRadius=20, radius=60, lineWidth=10, innerLineWidth=5)

    def test_radius(self):
        # Define some cases to test
        cases = [
            {"outer": 1, "inner": 0.5, "units": "height"},
            {"outer": 100, "inner": 50, "units": "pix"},
        ]

        for case in cases:
            # Set radiae (radiuses?) and units
            self.obj.units = case['units']
            self.obj.outerRadius = case['outer']
            self.obj.innerRadius = case['inner']
            # Check that the target's size is set to twice the radius value
            assert self.obj.outer._size == layout.Size(case['outer']*2, case['units'], self.win)
            assert self.obj.inner._size == layout.Size(case['inner']*2, case['units'], self.win)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""This module has tools for fetching data about the system or the current
Python process. Such info can be useful for understanding the context in which
an experiment was run.
"""

__all__ = [
    'RunTimeInfo',
    'getMemoryUsage',
    'getRAM',
    'APP_FLAG_LIST',  # user might want to see these
    'APP_IGNORE_LIST',
    # These should be hidden, but I'm unsure if somewhere `import *` is being
    # used so I'm adding them for now to prevent breakage. - mdc
    '_getUserNameUID',
    '_getHashGitHead',
    '_getSha1hexDigest',
    '_getHgVersion',
    '_getSvnVersion',
    '_thisProcess'
]

import sys
import os
import platform
import io

from pyglet.gl import gl_info, GLint, glGetIntegerv, GL_MAX_ELEMENTS_VERTICES
import numpy
import scipy
import matplotlib
import pyglet
try:
    import ctypes
    haveCtypes = True
except ImportError:
    haveCtypes = False
import hashlib
import wx
import locale
import subprocess
import psutil

from psychopy import visual, logging, core, data, web
from psychopy.core import shellCall
from psychopy.platform_specific import rush
from psychopy import __version__ as psychopyVersion

# List of applications to flag as problematic while running an experiment. These
# apps running in the background consume resources (CPU, GPU and memory) which
# may interfere with a PsychoPy experiment. If these apps are allowed to run, it
# may result in poor timing, glitches, dropped frames, etc.
#
# Names that appear here are historically known to affect performance. The user
# can check if these processes are running using `RunTimeInfo()` and shut them
# down. App names are matched in a case insensitive way from the start of the
# name string obtained from `psutils`.
APP_FLAG_LIST = [
    # web browsers can burn CPU cycles
    'Firefox',
    'Safari',
    'Explorer',
    'Netscape',
    'Opera',
    'Google Chrome',
    'Dropbox',
    'BitTorrent',
    'iTunes',  # but also matches iTunesHelper (add to ignore-list)
    'mdimport',
    'mdworker',
    'mds',  # can have high CPU
    # productivity
    'Office',
    'KeyNote',
    'Pages',
    'LaunchCFMApp',  # on mac, MS Office (Word etc) can be launched by this
    'Skype',
    'VirtualBox',
    'VBoxClient',  # virtual machine as host or client
    'Parallels',
    'Coherence',
    'prl_client_app',
    'prl_tools_service',
    'VMware'  # just a guess
    # gaming, may need to be started for VR support
    'Steam'
    'Oculus'
]

# Apps listed here will not be flagged if a partial match exist in
# `APP_FLAG_LIST`. This list is checked first before `RunTimeInfo()` looks for a
# name in `APP_FLAG_LIST`. You can also add names here to eliminate ambiguity,
# for instance if 'Dropbox' is in `APP_FLAG_LIST`, then `DropboxUpdate` will
# also be flagged. You can prevent this by adding 'DropboxUpdate' to
# `APP_IGNORE_LIST`.
APP_IGNORE_LIST = [
    # shells
    'ps',
    'login',
    '-tcsh',
    'bash',
    # helpers and updaters
    'iTunesHelper',
    'DropboxUpdate',
    'OfficeClickToRun'
]


class RunTimeInfo(dict):
    """Returns a snapshot of your configuration at run-time, for immediate or
    archival use.

    Returns a dict-like object with info about PsychoPy, your experiment script,
    the system & OS, your window and monitor settings (if any), python &
    packages, and openGL.

    If you want to skip testing the refresh rate, use 'refreshTest=None'

    Example usage: see runtimeInfo.py in coder demos.

    Parameters
    ----------
    win : :class:`~psychopy.visual.Window`, False or None
        What window to use for refresh rate testing (if any) and settings.
        `None` -> temporary window using defaults; `False` -> no window created,
        used, nor profiled; a `Window()` instance you have already created one.
    author : str or None
        `None` will try to autodetect first __author__ in sys.argv[0], whereas
        a `str` being user-supplied author info (of an experiment).
    version : str or None
        `None` try to autodetect first __version__ in sys.argv[0] or `str` being
        the user-supplied version info (of an experiment).
    verbose : bool
        Show additional information. Default is `False`.
    refreshTest : str, bool or None
        True or 'grating' = assess refresh average, median, and SD of 60
        win.flip()s, using visual.getMsPerFrame() 'grating' = show a visual
        during the assessment; `True` = assess without a visual. Default is
        `'grating'`.
    userProcsDetailed: bool
        Get details about concurrent user's processes (command, process-ID).
        Default is `False`.

    Returns
    -------
    A flat dict (but with several groups based on key names):

    psychopy : version, rush() availability
        psychopyVersion, psychopyHaveExtRush, git branch and current commit hash
        if available

    experiment : author, version, directory, name, current time-stamp, SHA1
        digest, VCS info (if any, svn or hg only),
        experimentAuthor, experimentVersion, ...

    system : hostname, platform, user login, count of users,
        user process info (count, cmd + pid), flagged processes
        systemHostname, systemPlatform, ...

    window : (see output; many details about the refresh rate, window,
        and monitor; units are noted)
        windowWinType, windowWaitBlanking, ...windowRefreshTimeSD_ms,
        ... windowMonitor.<details>, ...

    python : version of python, versions of key packages
        (wx, numpy, scipy, matplotlib, pyglet, pygame)
        pythonVersion, pythonScipyVersion, ...

    openGL : version, vendor, rendering engine, plus info on whether
        several extensions are present
        openGLVersion, ..., openGLextGL_EXT_framebuffer_object, ...

    """
    # Author: 2010 written by Jeremy Gray, input from Jon Peirce and
    # Alex Holcombe
    def __init__(self, author=None, version=None, win=None,
                 refreshTest='grating', userProcsDetailed=False,
                 verbose=False):
        # this will cause an object to be created with all the same methods as
        # a dict
        dict.__init__(self)

        self['psychopyVersion'] = psychopyVersion
        # NB: this looks weird, but avoids setting high-priority incidentally
        self['psychopyHaveExtRush'] = rush(False)
        d = os.path.abspath(os.path.dirname(__file__))
        githash = _getHashGitHead(d)  # should be .../psychopy/psychopy/
        if githash:
            self['psychopyGitHead'] = githash

        self._setExperimentInfo(author, version, verbose)
        self._setSystemInfo()  # current user, locale, other software
        self._setCurrentProcessInfo(verbose, userProcsDetailed)

        # need a window for frame-timing, and some openGL drivers want
        # a window open
        if win is None:  # make a temporary window, later close it
            win = visual.Window(
                fullscr=True, monitor="testMonitor", autoLog=False)
            refreshTest = 'grating'
            usingTempWin = True
        elif win != False:
            # we were passed a window instance, use it:
            usingTempWin = False
            self.winautoLog = win.autoLog
            win.autoLog = False
        else:  # don't want any window
            usingTempWin = False

        if win:
            self._setWindowInfo(win, verbose, refreshTest, usingTempWin)

        self['pythonVersion'] = sys.version.split()[0]
        if verbose:
            self._setPythonInfo()
            if win:
                self._setOpenGLInfo()
        if usingTempWin:
            win.close()  # close after doing openGL
        elif win != False:
            win.autoLog = self.winautoLog  # restore

    def _setExperimentInfo(self, author, version, verbose):
        """Auto-detect __author__ and __version__ in sys.argv[0] (= the
        # users's script)
        """
        if not author or not version:
            lines = ''
            if os.path.isfile(sys.argv[0]):
                with io.open(sys.argv[0], 'r', encoding='utf-8-sig') as f:
                    lines = f.read()
            if not author and '__author__' in lines:
                linespl = lines.splitlines()
                while linespl[0].find('__author__') == -1:
                    linespl.pop(0)
                auth = linespl[0]
                if len(auth) and '=' in auth:
                    try:
                        author = str(eval(auth[auth.find('=') + 1:]))
                    except Exception:
                        pass
            if not version and '__version__' in lines:
                linespl = lines.splitlines()
                while linespl[0].find('__version__') == -1:
                    linespl.pop(0)
                ver = linespl[0]
                if len(ver) and ver.find('=') > 0:
                    try:
                        version = str(eval(ver[ver.find('=') + 1:]))
                    except Exception:
                        pass

        if author or verbose:
            self['experimentAuthor'] = author
        if version or verbose:
            self['experimentAuthVersion'] = version

        # script identity & integrity information:
        self['experimentScript'] = os.path.basename(sys.argv[0])  # file name
        scriptDir = os.path.dirname(os.path.abspath(sys.argv[0]))
        self['experimentScript.directory'] = scriptDir
        # sha1 digest, text-format compatibility
        scriptPath = os.path.abspath(sys.argv[0])
        key = 'experimentScript.digestSHA1'
        self[key] = _getSha1hexDigest(scriptPath, isfile=True)
        # subversion revision?
        try:
            svnrev, last, url = _getSvnVersion(scriptPath)  # svn revision
            if svnrev:  # or verbose:
                self['experimentScript.svnRevision'] = svnrev
                self['experimentScript.svnRevLast'] = last
                self['experimentScript.svnRevURL'] = url
        except Exception:
            pass
        # mercurical revision?
        try:
            hgChangeSet = _getHgVersion(scriptPath)
            if hgChangeSet:  # or verbose:
                self['experimentScript.hgChangeSet'] = hgChangeSet
        except Exception:
            pass

        # when was this run?
        self['experimentRunTime.epoch'] = core.getAbsTime()
        fmt = "%Y_%m_%d %H:%M (Year_Month_Day Hour:Min)"
        self['experimentRunTime'] = data.getDateStr(format=fmt)

    def _setSystemInfo(self):
        """System info
        """
        # machine name
        self['systemHostName'] = platform.node()

        self['systemMemTotalRAM'], self['systemMemFreeRAM'] = getRAM()

        # locale information:
        # (None, None) -> str
        loc = '.'.join([str(x) for x in locale.getlocale()])
        if loc == 'None.None':
            loc = locale.setlocale(locale.LC_ALL, '')
        # == the locale in use, from OS or user-pref
        self['systemLocale'] = loc

        # platform name, etc
        if sys.platform in ['darwin']:
            OSXver, _junk, architecture = platform.mac_ver()
            platInfo = 'darwin ' + OSXver + ' ' + architecture
            # powerSource = ...
        elif sys.platform.startswith('linux'):
            platInfo = 'linux ' + platform.release()
            # powerSource = ...
        elif sys.platform in ['win32']:
            platInfo = 'windowsversion=' + repr(sys.getwindowsversion())
            # powerSource = ...
        else:
            platInfo = ' [?]'
            # powerSource = ...
        self['systemPlatform'] = platInfo
        # self['systemPowerSource'] = powerSource

        # count all unique people (user IDs logged in), and find current user
        # name & UID
        self['systemUser'], self['systemUserID'] = _getUserNameUID()
        try:
            users = shellCall("who -q").splitlines()[0].split()
            self['systemUsersCount'] = len(set(users))
        except Exception:
            self['systemUsersCount'] = False

        # when last rebooted?
        try:
            lastboot = shellCall("who -b").split()
            self['systemRebooted'] = ' '.join(lastboot[2:])
        except Exception:  # windows
            sysInfo = shellCall('systeminfo').splitlines()
            lastboot = [line for line in sysInfo if line.startswith(
                "System Up Time") or line.startswith("System Boot Time")]
            lastboot += ['[?]']  # put something in the list just in case
            self['systemRebooted'] = lastboot[0].strip()

        # R (and r2py) for stats:
        try:
            Rver = shellCall(["R", "--version"])
            Rversion = Rver.splitlines()[0]
            if Rversion.startswith('R version'):
                self['systemRavailable'] = Rversion.strip()
            try:
                import rpy2
                self['systemRpy2'] = rpy2.__version__
            except ImportError:
                pass
        except Exception:
            pass

        # encryption / security tools:
        try:
            vers, se = shellCall('openssl version', stderr=True)
            if se:
                vers = str(vers) + se.replace('\n', ' ')[:80]
            if vers.strip():
                self['systemSec.OpenSSLVersion'] = vers
        except Exception:
            pass
        try:
            so = shellCall(['gpg', '--version'])
            if so.find('GnuPG') > -1:
                self['systemSec.GPGVersion'] = so.splitlines()[0]
                _home = [line.replace('Home:', '').lstrip()
                         for line in so.splitlines()
                         if line.startswith('Home:')]
                self['systemSec.GPGHome'] = ''.join(_home)
        except Exception:
            pass
        try:
            import ssl
            self['systemSec.pythonSSL'] = True
        except ImportError:
            self['systemSec.pythonSSL'] = False

        # pyo for sound:
            import importlib.util
            if importlib.util.find_spec('pyo') is not None:
                self['systemPyoVersion'] = '-'

        # flac (free lossless audio codec) for google-speech:
        flacv = ''
        if sys.platform == 'win32':
            flacexe = 'C:\\Program Files\\FLAC\\flac.exe'
            if os.path.exists(flacexe):
                flacv = core.shellCall(flacexe + ' --version')
        else:
            flac, se = core.shellCall('which flac', stderr=True)
            if not se and flac and not flac.find('Command not found') > -1:
                flacv = core.shellCall('flac --version')
        if flacv:
            self['systemFlacVersion'] = flacv

        # detect internet access or fail quickly:
        # web.setupProxy() & web.testProxy(web.proxies)  # can be slow
        # to fail if there's no connection
        self['systemHaveInternetAccess'] = web.haveInternetAccess()
        if not self['systemHaveInternetAccess']:
            self['systemHaveInternetAccess'] = 'False (proxies not attempted)'

    def _setCurrentProcessInfo(self, verbose=False, userProcsDetailed=False):
        """What other processes are currently active for this user?
        """
        systemProcPsu = []             # found processes
        systemProcPsuFlagged = []      # processes which are flagged
        systemUserProcFlaggedPID = []  # PIDs of those processes
        # lower case these names for matching
        appFlagListLowerCase = [pn.lower() for pn in APP_FLAG_LIST]
        appIgnoreListLowerCase = [pn.lower() for pn in APP_IGNORE_LIST]

        # iterate over processes retrieved by psutil
        for proc in psutil.process_iter(attrs=None, ad_value=None):
            try:
                processFullName = proc.name()  # get process name
                processPid = proc.pid
                processName = processFullName.lower()  # use for matching only
            except (psutil.AccessDenied, psutil.NoSuchProcess,
                    psutil.ZombieProcess):
                continue  # skip iteration on exception

            # check if process is in ignore list, skip if so
            for appIgnore in appIgnoreListLowerCase:
                # case-insensitive match from the start of string
                if processName.startswith(appIgnore):
                    break
            else:
                # if we get here, the name isn't in the ignore list
                for appFlag in appFlagListLowerCase:
                    if processName.startswith(appFlag):
                        # append actual name and PID to output lists
                        systemProcPsuFlagged.append(processFullName)
                        systemUserProcFlaggedPID.append(processPid)
                        break

            systemProcPsu.append(processName)

        # add items to dictionary
        self['systemUserProcCount'] = len(systemProcPsu)
        self['systemUserProcFlagged'] = systemProcPsuFlagged

        # if the user wants more ...
        if verbose and userProcsDetailed:
            self['systemUserProcCmdPid'] = systemProcPsu  # is this right?
            self['systemUserProcFlaggedPID'] = systemUserProcFlaggedPID

        # CPU speed (will depend on system busy-ness)
        d = numpy.array(numpy.linspace(0., 1., 1000000))
        t0 = core.getTime()
        numpy.std(d)
        t = core.getTime() - t0
        del d
        self['systemTimeNumpySD1000000_sec'] = t

    def _setWindowInfo(self, win, verbose=False, refreshTest='grating',
                       usingTempWin=True):
        """Find and store info about the window: refresh rate,
        configuration info.
        """

        if refreshTest in ['grating', True]:
            wantVisual = bool(refreshTest == 'grating')
            a, s, m = visual.getMsPerFrame(win, nFrames=120,
                                           showVisual=wantVisual)
            self['windowRefreshTimeAvg_ms'] = a
            self['windowRefreshTimeMedian_ms'] = m
            self['windowRefreshTimeSD_ms'] = s
        if usingTempWin:
            return

        # These 'configuration lists' control what attributes are reported.
        # All desired attributes/properties need a legal internal name,
        # e.g., win.winType. If an attr is callable, its gets called with
        # no arguments, e.g., win.monitor.getWidth()
        winAttrList = ['winType', '_isFullScr', 'units',
                       'monitor', 'pos', 'screen', 'rgb', 'size']
        winAttrListVerbose = ['allowGUI', 'useNativeGamma',
                              'recordFrameIntervals', 'waitBlanking',
                              '_haveShaders', 'refreshThreshold']
        if verbose:
            winAttrList += winAttrListVerbose

        monAttrList = ['name', 'getDistance', 'getWidth', 'currentCalibName']
        monAttrListVerbose = ['getGammaGrid', 'getLinearizeMethod',
                              '_gammaInterpolator', '_gammaInterpolator2']
        if verbose:
            monAttrList += monAttrListVerbose
        if 'monitor' in winAttrList:
            # replace 'monitor' with all desired monitor.<attribute>
            # retain list-position info, put monitor stuff there
            i = winAttrList.index('monitor')
            del winAttrList[i]
            for monAttr in monAttrList:
                winAttrList.insert(i, 'monitor.' + monAttr)
                i += 1
        for winAttr in winAttrList:
            try:
                attrValue = eval('win.' + winAttr)
            except AttributeError:
                msg = ('AttributeError in RuntimeInfo._setWindowInfo(): '
                       'Window instance has no attribute')
                logging.warning(0, msg, winAttr)
                continue
            if hasattr(attrValue, '__call__'):
                try:
                    a = attrValue()
                    attrValue = a
                except Exception:
                    msg = ('Warning: could not get a value from win. '
                           '%s()  (expects arguments?)' % winAttr)
                    print(msg)
                    continue
            while winAttr[0] == '_':
                winAttr = winAttr[1:]
            winAttr = winAttr[0].capitalize() + winAttr[1:]
            winAttr = winAttr.replace('Monitor._', 'Monitor.')
            if winAttr in ('Pos', 'Size'):
                winAttr += '_pix'
            if winAttr in ('Monitor.getWidth', 'Monitor.getDistance'):
                winAttr += '_cm'
            if winAttr in ('RefreshThreshold'):
                winAttr += '_sec'
            self['window' + winAttr] = attrValue

    def _setPythonInfo(self):
        """External python packages, python details
        """
        self['pythonNumpyVersion'] = numpy.__version__
        self['pythonScipyVersion'] = scipy.__version__
        self['pythonWxVersion'] = wx.version()
        self['pythonMatplotlibVersion'] = matplotlib.__version__
        self['pythonPygletVersion'] = pyglet.version
        try:
            from pygame import __version__ as pygameVersion
        except ImportError:
            pygameVersion = '(no pygame)'
        self['pythonPygameVersion'] = pygameVersion

        # Python gory details:
        self['pythonFullVersion'] = sys.version.replace('\n', ' ')
        self['pythonExecutable'] = sys.executable

    def _setOpenGLInfo(self):
        # OpenGL info:
        self['openGLVendor'] = gl_info.get_vendor()
        self['openGLRenderingEngine'] = gl_info.get_renderer()
        self['openGLVersion'] = gl_info.get_version()
        GLextensionsOfInterest = ('GL_ARB_multitexture',
                                  'GL_EXT_framebuffer_object',
                                  'GL_ARB_fragment_program',
                                  'GL_ARB_shader_objects',
                                  'GL_ARB_vertex_shader',
                                  'GL_ARB_texture_non_power_of_two',
                                  'GL_ARB_texture_float', 'GL_STEREO')

        for ext in GLextensionsOfInterest:
            self['openGLext.' + ext] = bool(gl_info.have_extension(ext))

        maxVerts = GLint()
        glGetIntegerv(GL_MAX_ELEMENTS_VERTICES, maxVerts)
        self['openGLmaxVerticesInVertexArray'] = maxVerts.value

    def __repr__(self):
        """Return a string that is a legal python (dict), and close
        to YAML, .ini, and configObj syntax
        """
        info = '{\n#[ PsychoPy3 RuntimeInfoStart ]\n'
        sections = ['PsychoPy', 'Experiment',
                    'System', 'Window', 'Python', 'OpenGL']
        for sect in sections:
            info += '  #[[ %s ]] #---------\n' % (sect)
            sectKeys = [k for k in list(self.keys(
            )) if k.lower().find(sect.lower()) == 0]
            # get keys for items matching this section label;
            #  use reverse-alpha order if easier to read:
            revSet = ('PsychoPy', 'Window', 'Python', 'OpenGL')
            sectKeys.sort(reverse=bool(sect in revSet))
            for k in sectKeys:
                selfk = self[k]  # alter a copy for display purposes
                try:
                    if type(selfk) == type('abc'):
                        selfk = selfk.replace('"', '').replace('\n', ' ')
                    elif '_ms' in k:  # type(selfk) == type(0.123):
                        selfk = "%.3f" % selfk
                    elif '_sec' in k:
                        selfk = "%.4f" % selfk
                    elif '_cm' in k:
                        selfk = "%.1f" % selfk
                except Exception:
                    pass
                # then strcat unique proc names
                if (k in ('systemUserProcFlagged', 'systemUserProcCmdPid') and
                        selfk is not None and
                        len(selfk)):
                    prSet = []
                    for pr in self[k]:  # str -> list of lists
                        if ' ' in pr:  # add single quotes if file has spaces
                            pr = "'" + pr + "'"
                        # first item in sublist is proc name (CMD)
                        prSet += [pr]
                    selfk = ' '.join(list(set(prSet)))
                # suppress display PID info -- useful at run-time, never useful
                # in an archive
                if k != 'systemUserProcFlaggedPID':
                    info += '    "%s": "%s",\n' % (k, selfk)
        info += '#[ PsychoPy3 RuntimeInfoEnd ]\n}\n'
        return info

    def __str__(self):
        """Return a string intended for printing to a log file
        """
        infoLines = self.__repr__()
        # remove enclosing braces from repr
        info = infoLines.splitlines()[1:-1]
        for i, line in enumerate(info):
            if 'openGLext' in line:
                # swap order for OpenGL extensions -- much easier to read
                tmp = line.split(':')
                info[i] = ': '.join(['   ' + tmp[1].replace(',', ''),
                                     tmp[0].replace('    ', '') + ','])
            info[i] = info[i].rstrip(',')
        info = '\n'.join(info).replace('"', '') + '\n'
        return info


def _getHashGitHead(gdir='.'):
    if not os.path.isdir(gdir):
        raise OSError('not a directory')
    try:
        git_hash = subprocess.check_output('git rev-parse --verify HEAD',
                                           cwd=gdir,
                                           shell=True, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError:
        return None  # no git
    git_branches = subprocess.check_output('git branch', cwd=gdir, shell=True)
    git_branch = [line.split()[1] for line in git_branches.splitlines()
                  if line.startswith(b'*')]
    if len(git_branch):
        return "{} {}".format(git_branch[0], git_hash.strip())
    else:
        return '(unknown branch)'


def _getSvnVersion(filename):
    """Tries to discover the svn version (revision #) for a file.

    Not thoroughly tested; untested on Windows Vista, Win 7, FreeBSD

    :Author:
        - 2010 written by Jeremy Gray
    """
    if not (os.path.exists(filename) and
            os.path.isdir(os.path.join(os.path.dirname(filename), '.svn'))):
        return None, None, None
    svnRev, svnLastChangedRev, svnUrl = None, None, None
    if (sys.platform in ('darwin', 'freebsd') or
            sys.platform.startswith('linux')):
        try:
            # expects a filename, not dir
            svninfo = shellCall(['svn', 'info', filename])
        except Exception:
            svninfo = ''
        for line in svninfo.splitlines():
            if line.startswith('URL:'):
                svnUrl = line.split()[1]
            elif line.startswith('Revision: '):
                svnRev = line.split()[1]
            elif line.startswith('Last Changed Rev'):
                svnLastChangedRev = line.split()[3]
    else:
        # worked for me on Win XP sp2 with TortoiseSVN (SubWCRev.exe)
        try:
            stdout = shellCall(['subwcrev', filename])
        except Exception:
            stdout = ''
        for line in stdout.splitlines():
            if line.startswith('Last committed at revision'):
                svnRev = line.split()[4]
            elif line.startswith('Updated to revision'):
                svnLastChangedRev = line.split()[3]
    return svnRev, svnLastChangedRev, svnUrl


def _getHgVersion(filename):
    """Tries to discover the mercurial (hg) parent and id of a file.

    Not thoroughly tested; untested on Windows Vista, Win 7, FreeBSD

    :Author:
        - 2010 written by Jeremy Gray
    """
    dirname = os.path.dirname
    if (not os.path.exists(filename) or
            not os.path.isdir(os.path.join(dirname(filename), '.hg'))):
        return None
    try:
        hgParentLines, err = shellCall(['hg', 'parents', filename],
                                       stderr=True)
        changeset = hgParentLines.splitlines()[0].split()[-1]
    except Exception:
        changeset = ''
    try:
        hgID, err = shellCall(['hg', 'id', '-nibt', dirname(filename)],
                              stderr=True)
    except Exception:
        if err:
            hgID = ''

    if len(hgID) or len(changeset):
        return hgID.strip() + ' | parent: ' + changeset.strip()
    else:
        return None


def _getUserNameUID():
    """Return user name, UID.

    UID values can be used to infer admin-level:
    -1=undefined, 0=full admin/root,
    >499=assume non-admin/root (>999 on debian-based)

    :Author:
        - 2010 written by Jeremy Gray
    """
    user = os.environ.get('USER', None) or os.environ.get('USERNAME', None)
    if not user:
        return 'undefined', '-1'

    if sys.platform not in ['win32']:
        uid = shellCall('id -u')
    else:
        uid = '1000'
        if haveCtypes and ctypes.windll.shell32.IsUserAnAdmin():
            uid = '0'
    return str(user), int(uid)


def _getSha1hexDigest(thing, isfile=False):
    """Returns base64 / hex encoded sha1 digest of str(thing), or
    of a file contents. Return None if a file is requested but no such
    file exists

    :Author:
        - 2010 Jeremy Gray; updated 2011 to be more explicit,
        - 2012 to remove sha.new()

    >>> _getSha1hexDigest('1')
    '356a192b7913b04c54574d18c28d46e6395428ab'
    >>> _getSha1hexDigest(1)
    '356a192b7913b04c54574d18c28d46e6395428ab'
    """
    digester = hashlib.sha1()
    if isfile:
        filename = thing
        if os.path.isfile(filename):
            f = open(filename, 'rb')
            # check file size < available RAM first? or update in chunks?
            digester.update(f.read())
            f.close()
        else:
            return None
    else:
        digester.update(str(thing))
    return digester.hexdigest()


def getRAM():
    """Return system's physical RAM & available RAM, in M.
    """
    totalRAM, available = psutil.virtual_memory()[0:2]
    return totalRAM / 1048576., available / 1048576.


# faster to get the current process only once:
_thisProcess = psutil.Process()


def getMemoryUsage():
    """Get the memory (RAM) currently used by this Python process, in M.
    """
    return _thisProcess.memory_info()[0] / 1048576.



from psychopy.sound._base import apodize, HammingWindow
from psychopy.constants import FINISHED
from psychopy.exceptions import DependencyError
import numpy as np
import pytest

import psychopy.sound.backend_ptb as ptb

"""
We need to test that the new block-by-block hamming works the same as the
(simpler) method of adding the hamming window to the initial complete array
(using the apodize function)
"""

sampleRate = 44100
thisFreq = 100
secs = 0.3
nSamples = int(secs * sampleRate)
t = np.arange(0.0, 1.0, 1.0 / nSamples)*secs
sndArray = np.sin(t*2*np.pi*thisFreq)

plotting = False
if plotting:
    import matplotlib.pyplot as plt


@pytest.mark.needs_sound
def test_HammingSmallBlock():
    blockSize = 64
    snd1 = apodize(sndArray, sampleRate)  # is 5 ms
    sndDev = ptb.SoundPTB(thisFreq, sampleRate=sampleRate, secs=secs,
                          hamming=True, blockSize=blockSize)
    snd2 = []
    while sndDev.status != FINISHED:
        block = sndDev._nextBlock()
        snd2.extend(block)
    snd2 = np.array(snd2)

    if plotting:
        plt.subplot(2,1,1)
        plt.plot(snd1, 'b-')
        plt.plot(snd2, 'r--')
        plt.subplot(2,1,2)
        plt.plot(t, snd2[0:sampleRate*secs]-snd1)
        plt.show()



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
import sys
import os
import copy
import inspect
import warnings
import numpy
import numbers  # numbers.Integral is like (int, long) but supports Py3
import datetime
from ..errors import print2err
import re
import collections.abc

########################
#
# .yaml read / write

try:
    from yaml import load as yload
    from yaml import dump as ydump
    from yaml import CLoader as yLoader, CDumper as yDumper
except ImportError:
    from yaml import Loader as yLoader, Dumper as yDumper

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable


def saveConfig(config, dst_path):
    '''
    Saves a config dict to dst_path in YAML format.
    '''
    ydump(config, open(dst_path, 'w'), Dumper=yDumper)
    return os.path.exists(dst_path)


def readConfig(scr_path):
    '''
    Returns the config dict loaded from scr_path, which must be the path to
    a YAML file.
    '''
    return yload(open(scr_path, 'r'), Loader=yLoader)

def mergeConfigurationFiles(base_config_file_path, update_from_config_file_path, merged_save_to_path):
    """Merges two iohub configuration files into one and saves it to a file
    using the path/file name in merged_save_to_path."""
    base_config = yload(open(base_config_file_path, 'r'), Loader=yLoader)
    update_from_config = yload(
        open(
            update_from_config_file_path,
            'r'),
        Loader=yLoader)

    def merge(update, base):
        if isinstance(update, dict) and isinstance(base, dict):
            for k, v in base.items():
                if k not in update:
                    update[k] = v
                else:
                    if isinstance(update[k], list):
                        if isinstance(v, list):
                            v.extend(update[k])
                            update[k] = v
                        else:
                            update[k].insert(0, v)
                    else:
                        update[k] = merge(update[k], v)
        return update

    merged = merge(copy.deepcopy(update_from_config), base_config)
    ydump(merged, open(merged_save_to_path, 'w'), Dumper=yDumper)

    return merged
########################


def normjoin(*path_parts):
    """
    normjoin combines the following Python os.path functions in the following
    call order:
        * join
        * normcase
        * normpath

    Args:
        *path_parts (tuple): The tuple of path parts to pass to os.path.join.

    Returns:

    """
    return os.path.normpath(os.path.normcase(os.path.join(*path_parts)))


def addDirectoryToPythonPath(path_from_iohub_root, leaf_folder=''):
    from .. import IOHUB_DIRECTORY
    dir_path = os.path.join(
        IOHUB_DIRECTORY,
        path_from_iohub_root,
        sys.platform,
        'python{0}{1}'.format(
            *
            sys.version_info[
                0:2]),
        leaf_folder)
    if os.path.isdir(dir_path) and dir_path not in sys.path:
        sys.path.append(dir_path)
    else:
        print2err('Could not add path: ', dir_path)
        dir_path = None
    return dir_path


def module_path(local_function):
    """returns the module path without the use of __file__.

    Requires a function defined
    locally in the module. from http://stackoverflow.com/questions/729583/getting-file-path-of-imported-module

    """
    return os.path.abspath(inspect.getsourcefile(local_function))


def module_directory(local_function):
    mp = module_path(local_function)
    moduleDirectory, mname = os.path.split(mp)
    return moduleDirectory


def isIterable(o):
    return isinstance(o, Iterable)


# Get available device module paths
def getDevicePaths(device_name=""):
    """
    """
    from psychopy.iohub.devices import import_device
    iohub_device_path = module_directory(import_device)
    if device_name:
        iohub_device_path = os.path.join(iohub_device_path, device_name.replace('.', os.path.sep))
    scs_yaml_paths = []
    for root, dirs, files in os.walk(iohub_device_path):
        device_folder = None
        for file in files:
            if file == 'supported_config_settings.yaml':
                device_folder = root
                break
        if device_folder:
            for dfile in files:
                if dfile.startswith("default_") and dfile.endswith('.yaml'):
                    scs_yaml_paths.append((device_folder, dfile))
    return scs_yaml_paths

def getDeviceDefaultConfig(device_name, builder_hides=True):
    """
    Return the default iohub config dictionary for the given device(s). The dictionary contains the
    (possibly nested) settings that should be displayed for the device (the dict item key) and the default value
    (the dict item value).

    Example:
        import pprint
        gp3_et_conf_defaults = getDeviceDefaultConfig('eyetracker.hw.gazepoint.gp3')
        pprint.pprint(gp3_et_conf_defaults)

    Output:
        {'calibration': {'target_delay': 0.5, 'target_duration': 1.25},
         'event_buffer_length': 1024,
         'manufacturer_name': 'GazePoint',
         'model_name': 'GP3',
         'monitor_event_types': ['BinocularEyeSampleEvent',
                                 'FixationStartEvent',
                                 'FixationEndEvent'],
         'network_settings': {'ip_address': '127.0.0.1', 'port': 4242},
         'runtime_settings': {'sampling_rate': 60},
         'save_events': True,
         'stream_events': True}
    """
    if device_name.endswith(".EyeTracker"):
        device_name = device_name[:-11]
    device_paths = getDevicePaths(device_name)
    device_configs = []
    for dpath, dconf in device_paths:
        dname, dconf_dict = list(readConfig(os.path.join(dpath, dconf)).items())[0]
        if builder_hides:
            to_hide = dconf_dict.get('builder_hides', [])
            for param in to_hide:
                if param.find('.') >= 0:
                    # it is a nested param
                    param_tokens = param.split('.')
                    cdict = dconf_dict
                    for pt in param_tokens[:-1]:
                        cdict = cdict.get(pt)
                    try:
                        del cdict[param_tokens[-1]]
                    except KeyError:
                        # key does not exist
                        pass
                    except TypeError:
                        pass
                else:
                    del dconf_dict[param]
        device_configs.append({dname: dconf_dict})
    if len(device_configs) == 1:
        # simplify return value when only one device was requested
        return list(device_configs[0].values())[0]
    return device_configs

def getDeviceNames(device_name="eyetracker.hw", get_paths=True):
    """
    Return a list of iohub eye tracker device names, as would be used as keys to launchHubServer. If get_paths is true,
    return both device manufacturer name (for display in builder) as well as iohub device name.

    Example:
        eyetrackers = getDeviceNames()
        print(eyetrackers)

    Output:
        [('GazePoint', 'eyetracker.hw.gazepoint.gp3.EyeTracker'),
         ('MouseGaze', 'eyetracker.hw.mouse.EyeTracker'),
         ('SR Research Ltd', 'eyetracker.hw.sr_research.eyelink.EyeTracker'),
         ('Tobii Technology', 'eyetracker.hw.tobii.EyeTracker')]
    """
    names = []
    dconfigs = getDeviceDefaultConfig(device_name)
    for dcfg in dconfigs:
        d_path = tuple(dcfg.keys())[0]
        d_config = tuple(dcfg.values())[0]
        if get_paths is False:
            names.append(d_path)
        else:
            names.append((d_config.get('manufacturer_name'), d_path))
    return names

def getDeviceFile(device_name, file_name):
    """
    Returns the contents of file_name for the specified device. If file_name does not exist, None is returned.

    :param device_name: iohub device name
    :param: file_name: name of device yaml file to load
    :return: dict
    """
    if device_name.endswith(".EyeTracker"):
        device_name = device_name[:-11]
    device_paths = getDevicePaths(device_name)
    device_sconfigs = []
    for dpath, _ in device_paths:
        device_sconfigs.append(readConfig(os.path.join(dpath, file_name)))
    if len(device_sconfigs) == 1:
        # simplify return value when only one device was requested
        return list(device_sconfigs[0].values())[0]
    return device_sconfigs

def getDeviceSupportedConfig(device_name):
    """
    Returns the contents of the supported_config_settings.yaml for the specified device.

    :param device_name: iohub device name
    :return: dict
    """
    return getDeviceFile(device_name, 'supported_config_settings.yaml')

if sys.platform == 'win32':
    import pythoncom

    def win32MessagePump():
        """Pumps the Windows Message Queue so that PsychoPy
        Window(s) lock up if psychopy has not called
        the windows 'dispatch_events()' method recently.

        If you are not flipping regularly (say because you do not need
        to and do not want to block frequently, you can call this, which
        will not block waiting for messages, but only pump out what is
        in the queue already. On an i7 desktop, this call method takes
        between 10 and 90 usec.

        """
        if pythoncom.PumpWaitingMessages() == 1:
            raise KeyboardInterrupt()
else:
    def win32MessagePump():
        pass

# PsychoPy Window Hide / Show functions.
# Windows 10 and macOS have different code that needs to be called
# to show a second full screen window on top of an existing one, like
# is done by most eye tracker calibration routines.
def hideWindow(win, force=False):
    """
    If needed, hide / minimize the in.
    :param win: PsychoPy window instance
    :return: None
    """
    if force or sys.platform == 'win32':
        if win._isFullScr:
            win.winHandle.minimize()  # minimize the PsychoPy window
            win.winHandle.set_fullscreen(False)
    elif sys.platform == 'darwin':
        pass
    elif sys.platform == 'linux':
        # TODO: test on Linux, assuming same as macOS right now
        pass
    else:
        print("Warning: Unhandled sys.platform: ", sys.platform)

def showWindow(win, force=False):
    """
    If needed, hide / minimize the in.
    :param win: PsychoPy window instance
    :return: None
    """
    if force or sys.platform == 'win32':
        if win._isFullScr:
            win.winHandle.set_fullscreen(True)
            win.winHandle.maximize()  # maximize the PsychoPy window
    elif sys.platform == 'darwin':
        pass
    elif sys.platform == 'linux':
        # TODO: test on Linux, assuming same as macOS right now
        pass
    else:
        print("Warning: Unhandled sys.platform: ", sys.platform)

# Recursive updating of values from one dict into another if the key does not key exist.
# Supported nested dicts and uses deep copy when setting values in the
# target dict.


def updateDict(add_to, add_from):
    for key, value in add_from.items():
        if key not in add_to:
            add_to[key] = copy.deepcopy(value)
        elif isinstance(value, dict) and isinstance(add_to[key], dict):
            updateDict(add_to[key], value)

def updateSettings(d, u):
    for k, v in u.items():
        if type(k) == bytes:
            k = k.decode('UTF-8')
        if isinstance(v, collections.abc.Mapping):
            d[k] = updateSettings(d.get(k, {}), v)
        else:
            if type(v) == bytes:
                v = v.decode('UTF-8')
            d[k] = v
    return d

# Convert Camel to Snake variable name format
first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')


def convertCamelToSnake(name, lower_snake=True):
    s1 = first_cap_re.sub(r'\1_\2', name)
    if lower_snake:
        return all_cap_re.sub(r'\1_\2', s1).lower()
    return all_cap_re.sub(r'\1_\2', s1).upper()


# A couple date / time related utility functions

getCurrentDateTime = datetime.datetime.now
getCurrentDateTimeString = lambda: getCurrentDateTime().strftime("%Y-%m-%d %H:%M")


# rgb255 color utils
def hilo(a, b, c):
    if c < b:
        b, c = c, b
    if b < a:
        a, b = b, a
    if c < b:
        b, c = c, b
    return a + c


def complement(r, g, b):
    if r == g == b:
        # handle mono color
        if r >= 128:
            return 0, 0, 0
        return 255, 255, 255
    k = hilo(r, g, b)
    return tuple(k - u for u in (r, g, b))

class NumPyRingBuffer():
    """NumPyRingBuffer is a circular buffer implemented using a one dimensional
    numpy array on the backend. The algorithm used to implement the ring buffer
    behavior does not require any array copies to occur while the ring buffer
    is maintained, while at the same time allowing sequential element access
    into the numpy array using a subset of standard slice notation.

    When the circular buffer is created, a maximum size , or maximum
    number of elements,  that the buffer can hold *must* be specified. When
    the buffer becomes full, each element added to the buffer removes the oldest
    element from the buffer so that max_size is never exceeded.

    Items are added to the ring buffer using the classes append method.

    The current number of elements in the buffer can be retrieved using the
    getLength() method of the class.

    The isFull() method can be used to determine if
    the ring buffer has reached its maximum size, at which point each new element
    added will disregard the oldest element in the array.

    The getElements() method is used to retrieve the actual numpy array containing
    the elements in the ring buffer. The element in index 0 is the oldest remaining
    element added to the buffer, and index n (which can be up to max_size-1)
    is the the most recent element added to the buffer.

    Methods that can be called from a standard numpy array can also be called using the
    NumPyRingBuffer instance created. However Numpy module level functions will not accept
    a NumPyRingBuffer as a valid argument.

    To clear the ring buffer and start with no data in the buffer, without
    needing to create a new NumPyRingBuffer object, call the clear() method
    of the class.

    Example::

        ring_buffer=NumPyRingBuffer(10)

        for i in range(25):
            ring_buffer.append(i)
            print('-------')
            print('Ring Buffer Stats:')
            print('\tWindow size: ',len(ring_buffer))
            print('\tMin Value: ',ring_buffer.min())
            print('\tMax Value: ',ring_buffer.max())
            print('\tMean Value: ',ring_buffer.mean())
            print('\tStandard Deviation: ',ring_buffer.std())
            print('\tFirst 3 Elements: ',ring_buffer[:3])
            print('\tLast 3 Elements: ',ring_buffer[-3:])

    """

    def __init__(self, max_size, dtype=numpy.float32):
        self._dtype = dtype
        self._npa = numpy.empty(max_size * 2, dtype=dtype)
        self.max_size = max_size
        self._index = 0

    def append(self, element):
        """Add element e to the end of the RingBuffer. The element must match
        the numpy data type specified when the NumPyRingBuffer was created. By
        default, the RingBuffer uses float32 values.

        If the Ring Buffer is full, adding the element to the end of the array
        removes the currently oldest element from the start of the array.

        :param numpy.dtype element: An element to add to the RingBuffer.
        :returns None:

        """
        i = self._index
        self._npa[i % self.max_size] = element
        self._npa[(i % self.max_size) + self.max_size] = element
        self._index += 1

    def getElements(self):
        """Return the numpy array being used by the RingBuffer, the length of
        which will be equal to the number of elements added to the list, or the
        last max_size elements added to the list. Elements are in order of
        addition to the ring buffer.

        :param None:
        :returns numpy.array: The array of data elements that make up the Ring Buffer.

        """
        return self._npa[
            self._index %
            self.max_size:(
                self._index %
                self.max_size) + self.max_size]

    def isFull(self):
        """Indicates if the RingBuffer is at it's max_size yet.

        :param None:
        :returns bool: True if max_size or more elements have been added to the RingBuffer; False otherwise.

        """
        return self._index >= self.max_size

    def clear(self):
        """Clears the RingBuffer. The next time an element is added to the
        buffer, it will have a size of one.

        :param None:
        :returns None:

        """
        self._index = 0

    def __setitem__(self, indexs, v):
        if isinstance(indexs, (list, tuple)):
            for i in indexs:
                if isinstance(i, numbers.Integral):
                    i = i + self._index
                    self._npa[i % self.max_size] = v
                    self._npa[(i % self.max_size) + self.max_size] = v
                elif isinstance(i, slice):
                    istart = indexs.start
                    if istart is None:
                        istart = 0
                    istop = indexs.stop
                    if indexs.stop is None:
                        istop = 0
                    start = istart + self._index
                    stop = istop + self._index
                    self._npa[
                        slice(
                            start %
                            self.max_size,
                            stop %
                            self.max_size,
                            i.step)] = v
                    self._npa[
                        slice(
                            (start %
                             self.max_size) + self.max_size, (stop %
                                                              self.max_size) + self.max_size, i.step)] = v
        elif isinstance(indexs, numbers.Integral):
            i = indexs + self._index
            self._npa[i % self.max_size] = v
            self._npa[(i % self.max_size) + self.max_size] = v
        elif isinstance(indexs, slice):
            istart = indexs.start
            if istart is None:
                istart = 0
            istop = indexs.stop
            if indexs.stop is None:
                istop = 0
            start = istart + self._index
            stop = istop + self._index
            self._npa[
                slice(
                    start %
                    self.max_size,
                    stop %
                    self.max_size,
                    indexs.step)] = v
            self._npa[
                slice(
                    (start %
                     self.max_size) +
                    self.max_size,
                    (stop %
                     self.max_size) +
                    self.max_size,
                    indexs.step)] = v
        else:
            raise TypeError()

    def __getitem__(self, indexs):
        current_array = self.getElements()
        if isinstance(indexs, (list, tuple)):
            rarray = []
            for i in indexs:
                if isinstance(i, int):
                    rarray.append(current_array[i])
                elif isinstance(i, slice):
                    rarray.extend(current_array[i])
            return numpy.asarray(rarray, dtype=self._dtype)
        elif isinstance(indexs, (int, slice)):
            return current_array[indexs]
        else:
            raise TypeError()

    def __getattr__(self, a):
        if self._index < self.max_size:
            return getattr(self._npa[:self._index], a)
        return getattr(
            self._npa[
                self._index %
                self.max_size:(
                    self._index %
                    self.max_size) + self.max_size], a)

    def __len__(self):
        if self.isFull():
            return self.max_size
        return self._index

###############################################################################
#
# Generate a set of points in a NxM grid. Useful for creating calibration target positions,
# or grid spaced fixation point positions that can be used for validation / fixation accuracy.
#


def generatedPointGrid(pixel_width, pixel_height, width_scalar=1.0,
                       height_scalar=1.0, horiz_points=5, vert_points=5):

    swidth = pixel_width * width_scalar
    sheight = pixel_height * height_scalar

    # center 0 on screen center
    x, y = numpy.meshgrid(numpy.linspace(-swidth / 2.0, swidth / 2.0, horiz_points),
                          numpy.linspace(-sheight / 2.0, sheight / 2.0, vert_points))
    points = numpy.column_stack((x.flatten(), y.flatten()))

    return points


# Rotate a set of points in 2D
#
# Rotate a set of n 2D points in the form [[x1,x1],[x2,x2],...[xn,xn]]
# around the 2D point origin (x0,y0), by ang radians.
# Returns the rotated point list.
#
# FROM:
# http://gis.stackexchange.com/questions/23587/how-do-i-rotate-the-polygon-about-an-anchor-point-using-python-script


def rotate2D(pts, origin, ang=None):
    '''pts = {} Rotates points(nx2) about center cnt(2) by angle ang(1) in radian'''
    if ang is None:
        ang = numpy.pi / 4
    return numpy.dot(pts - origin,
                     numpy.array([[numpy.cos(ang),
                                   numpy.sin(ang)],
                                  [-numpy.sin(ang),
                                   numpy.cos(ang)]])) + origin



# -*- coding: utf-8 -*-

###########################################################################
## Python code generated with wxFormBuilder (version 3.10.0-4761b0c)
## http://www.wxformbuilder.org/
##
## PLEASE DO *NOT* EDIT THIS FILE!
###########################################################################

import wx
import wx.xrc

###########################################################################
## Class ColorPickerDialog
###########################################################################

class ColorPickerDialog ( wx.Dialog ):

	def __init__( self, parent ):
		wx.Dialog.__init__ ( self, parent, id = wx.ID_ANY, title = u"Color Picker", pos = wx.DefaultPosition, size = wx.Size( 640,480 ), style = wx.DEFAULT_DIALOG_STYLE )

		self.SetSizeHints( wx.Size( -1,-1 ), wx.DefaultSize )

		szrMain = wx.BoxSizer( wx.VERTICAL )

		self.pnlColorSelector = wx.Panel( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.TAB_TRAVERSAL )
		szrColorSelector = wx.BoxSizer( wx.HORIZONTAL )

		self.pnlColorPreview = wx.Panel( self.pnlColorSelector, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.TAB_TRAVERSAL )
		self.pnlColorPreview.SetBackgroundColour( wx.SystemSettings.GetColour( wx.SYS_COLOUR_HIGHLIGHT ) )
		self.pnlColorPreview.SetMaxSize( wx.Size( 120,-1 ) )

		szrColorSelector.Add( self.pnlColorPreview, 1, wx.EXPAND|wx.LEFT|wx.RIGHT|wx.TOP, 5 )

		self.nbColorSelector = wx.Notebook( self.pnlColorSelector, wx.ID_ANY, wx.DefaultPosition, wx.Size( -1,-1 ), 0 )
		self.pnlRGBPage = wx.Panel( self.nbColorSelector, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.TAB_TRAVERSAL )
		szrRGBPage = wx.BoxSizer( wx.VERTICAL )

		fraRGBChannels = wx.StaticBoxSizer( wx.StaticBox( self.pnlRGBPage, wx.ID_ANY, u" RGB Channels " ), wx.VERTICAL )

		szrRGBChannels = wx.FlexGridSizer( 3, 3, 5, 10 )
		szrRGBChannels.AddGrowableCol( 1 )
		szrRGBChannels.SetFlexibleDirection( wx.BOTH )
		szrRGBChannels.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )

		self.lblRedChannel = wx.StaticText( fraRGBChannels.GetStaticBox(), wx.ID_ANY, u"R:", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lblRedChannel.Wrap( -1 )

		szrRGBChannels.Add( self.lblRedChannel, 0, wx.ALL|wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL, 0 )

		self.sldRedChannel = wx.Slider( fraRGBChannels.GetStaticBox(), wx.ID_ANY, 50, 0, 100, wx.DefaultPosition, wx.DefaultSize, wx.SL_HORIZONTAL )
		szrRGBChannels.Add( self.sldRedChannel, 0, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_VERTICAL, 0 )

		self.spnRedChannel = wx.SpinCtrlDouble( fraRGBChannels.GetStaticBox(), wx.ID_ANY, u"0.0", wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS, -1, 1, 0.000000, 0.01 )
		self.spnRedChannel.SetDigits( 4 )
		szrRGBChannels.Add( self.spnRedChannel, 0, wx.ALL|wx.EXPAND, 0 )

		self.lblGreenChannel = wx.StaticText( fraRGBChannels.GetStaticBox(), wx.ID_ANY, u"G:", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lblGreenChannel.Wrap( -1 )

		szrRGBChannels.Add( self.lblGreenChannel, 0, wx.ALL|wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL, 0 )

		self.sldGreenChannel = wx.Slider( fraRGBChannels.GetStaticBox(), wx.ID_ANY, 50, 0, 100, wx.DefaultPosition, wx.DefaultSize, wx.SL_HORIZONTAL )
		szrRGBChannels.Add( self.sldGreenChannel, 0, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_VERTICAL, 0 )

		self.spnGreenChannel = wx.SpinCtrlDouble( fraRGBChannels.GetStaticBox(), wx.ID_ANY, u"0.0", wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS, -1, 1, 0.000000, 0.01 )
		self.spnGreenChannel.SetDigits( 4 )
		szrRGBChannels.Add( self.spnGreenChannel, 0, wx.ALL|wx.EXPAND, 0 )

		self.lblBlueChannel = wx.StaticText( fraRGBChannels.GetStaticBox(), wx.ID_ANY, u"B:", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lblBlueChannel.Wrap( -1 )

		szrRGBChannels.Add( self.lblBlueChannel, 0, wx.ALL|wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL, 0 )

		self.sldBlueChannel = wx.Slider( fraRGBChannels.GetStaticBox(), wx.ID_ANY, 50, 0, 100, wx.DefaultPosition, wx.DefaultSize, wx.SL_HORIZONTAL )
		szrRGBChannels.Add( self.sldBlueChannel, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL|wx.EXPAND, 0 )

		self.spnBlueChannel = wx.SpinCtrlDouble( fraRGBChannels.GetStaticBox(), wx.ID_ANY, u"0.0", wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS, -1, 1, 0, 0.01 )
		self.spnBlueChannel.SetDigits( 4 )
		szrRGBChannels.Add( self.spnBlueChannel, 0, wx.ALL|wx.EXPAND, 0 )


		fraRGBChannels.Add( szrRGBChannels, 1, wx.EXPAND, 5 )


		szrRGBPage.Add( fraRGBChannels, 0, wx.ALL|wx.EXPAND, 5 )

		szrLowerRGBPage = wx.BoxSizer( wx.HORIZONTAL )

		fraHexRGB = wx.StaticBoxSizer( wx.StaticBox( self.pnlRGBPage, wx.ID_ANY, u"Hex/HTML" ), wx.VERTICAL )

		self.txtHexRGB = wx.TextCtrl( fraHexRGB.GetStaticBox(), wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
		fraHexRGB.Add( self.txtHexRGB, 0, wx.ALL|wx.EXPAND, 5 )


		szrLowerRGBPage.Add( fraHexRGB, 1, wx.ALL, 5 )

		fraRGBFormat = wx.StaticBoxSizer( wx.StaticBox( self.pnlRGBPage, wx.ID_ANY, u"RGB Format" ), wx.VERTICAL )

		self.rdoRGBModePsychoPy = wx.RadioButton( fraRGBFormat.GetStaticBox(), wx.ID_ANY, u"PsychoPy RGB [-1:1]", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.rdoRGBModePsychoPy.SetValue( True )
		fraRGBFormat.Add( self.rdoRGBModePsychoPy, 0, wx.ALL, 2 )

		self.rdoRGBModeNormalized = wx.RadioButton( fraRGBFormat.GetStaticBox(), wx.ID_ANY, u"Normalized RGB [0:1]", wx.DefaultPosition, wx.DefaultSize, 0 )
		fraRGBFormat.Add( self.rdoRGBModeNormalized, 0, wx.ALL, 2 )

		self.rdoRGBMode255 = wx.RadioButton( fraRGBFormat.GetStaticBox(), wx.ID_ANY, u"8-Bit RGB [0:255]", wx.DefaultPosition, wx.DefaultSize, 0 )
		fraRGBFormat.Add( self.rdoRGBMode255, 0, wx.ALL, 2 )


		szrLowerRGBPage.Add( fraRGBFormat, 1, wx.BOTTOM|wx.RIGHT|wx.TOP, 5 )


		szrRGBPage.Add( szrLowerRGBPage, 1, wx.EXPAND, 5 )


		self.pnlRGBPage.SetSizer( szrRGBPage )
		self.pnlRGBPage.Layout()
		szrRGBPage.Fit( self.pnlRGBPage )
		self.nbColorSelector.AddPage( self.pnlRGBPage, u"RGB", True )
		self.pnlHSVPage = wx.Panel( self.nbColorSelector, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.TAB_TRAVERSAL )
		szrHSVPage = wx.BoxSizer( wx.VERTICAL )

		fraHSVChannels = wx.StaticBoxSizer( wx.StaticBox( self.pnlHSVPage, wx.ID_ANY, u" HSV Channels " ), wx.VERTICAL )

		szrHSVChannels = wx.FlexGridSizer( 3, 3, 5, 10 )
		szrHSVChannels.AddGrowableCol( 1 )
		szrHSVChannels.SetFlexibleDirection( wx.BOTH )
		szrHSVChannels.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )

		self.lblHueChannel = wx.StaticText( fraHSVChannels.GetStaticBox(), wx.ID_ANY, u"H:", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lblHueChannel.Wrap( -1 )

		szrHSVChannels.Add( self.lblHueChannel, 0, wx.ALL|wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL, 0 )

		self.sldHueChannel = wx.Slider( fraHSVChannels.GetStaticBox(), wx.ID_ANY, 50, 0, 100, wx.DefaultPosition, wx.DefaultSize, wx.SL_HORIZONTAL )
		szrHSVChannels.Add( self.sldHueChannel, 0, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_VERTICAL, 0 )

		self.spnHueChannel = wx.SpinCtrlDouble( fraHSVChannels.GetStaticBox(), wx.ID_ANY, u"0.0", wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS, 0, 360, 0, 1 )
		self.spnHueChannel.SetDigits( 0 )
		szrHSVChannels.Add( self.spnHueChannel, 0, wx.ALL|wx.EXPAND, 0 )

		self.lblSaturationChannel = wx.StaticText( fraHSVChannels.GetStaticBox(), wx.ID_ANY, u"S:", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lblSaturationChannel.Wrap( -1 )

		szrHSVChannels.Add( self.lblSaturationChannel, 0, wx.ALL|wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL, 0 )

		self.sldStaturationChannel = wx.Slider( fraHSVChannels.GetStaticBox(), wx.ID_ANY, 50, 0, 100, wx.DefaultPosition, wx.DefaultSize, wx.SL_HORIZONTAL )
		szrHSVChannels.Add( self.sldStaturationChannel, 0, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_VERTICAL, 0 )

		self.spnSaturationChannel = wx.SpinCtrlDouble( fraHSVChannels.GetStaticBox(), wx.ID_ANY, u"0.0", wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS, -1, 1, 0.000000, 0.01 )
		self.spnSaturationChannel.SetDigits( 4 )
		szrHSVChannels.Add( self.spnSaturationChannel, 0, wx.ALL|wx.EXPAND, 0 )

		self.lblValueChannel = wx.StaticText( fraHSVChannels.GetStaticBox(), wx.ID_ANY, u"V:", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lblValueChannel.Wrap( -1 )

		szrHSVChannels.Add( self.lblValueChannel, 0, wx.ALL|wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL, 0 )

		self.sldValueChannel = wx.Slider( fraHSVChannels.GetStaticBox(), wx.ID_ANY, 50, 0, 100, wx.DefaultPosition, wx.DefaultSize, wx.SL_HORIZONTAL )
		szrHSVChannels.Add( self.sldValueChannel, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL|wx.EXPAND, 0 )

		self.spnValueChannel = wx.SpinCtrlDouble( fraHSVChannels.GetStaticBox(), wx.ID_ANY, u"0.0", wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS, -1, 1, 0.000000, 0.001 )
		self.spnValueChannel.SetDigits( 4 )
		szrHSVChannels.Add( self.spnValueChannel, 0, wx.ALL|wx.EXPAND, 0 )


		fraHSVChannels.Add( szrHSVChannels, 0, wx.EXPAND, 5 )


		szrHSVPage.Add( fraHSVChannels, 0, wx.ALL|wx.EXPAND, 5 )


		self.pnlHSVPage.SetSizer( szrHSVPage )
		self.pnlHSVPage.Layout()
		szrHSVPage.Fit( self.pnlHSVPage )
		self.nbColorSelector.AddPage( self.pnlHSVPage, u"HSV", False )

		szrColorSelector.Add( self.nbColorSelector, 1, wx.EXPAND|wx.TOP, 5 )

		lstColorPresetsChoices = []
		self.lstColorPresets = wx.ListBox( self.pnlColorSelector, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, lstColorPresetsChoices, 0|wx.ALWAYS_SHOW_SB )
		self.lstColorPresets.SetMinSize( wx.Size( 140,-1 ) )

		szrColorSelector.Add( self.lstColorPresets, 1, wx.EXPAND|wx.LEFT|wx.RIGHT|wx.TOP, 5 )


		self.pnlColorSelector.SetSizer( szrColorSelector )
		self.pnlColorSelector.Layout()
		szrColorSelector.Fit( self.pnlColorSelector )
		szrMain.Add( self.pnlColorSelector, 1, wx.EXPAND, 0 )

		self.stlMain = wx.StaticLine( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.LI_HORIZONTAL )
		szrMain.Add( self.stlMain, 0, wx.EXPAND |wx.ALL, 5 )

		self.pnlDlgButtons = wx.Panel( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.TAB_TRAVERSAL )
		szrDlgButtons = wx.FlexGridSizer( 0, 6, 0, 0 )
		szrDlgButtons.AddGrowableCol( 1 )
		szrDlgButtons.AddGrowableCol( 2 )
		szrDlgButtons.SetFlexibleDirection( wx.BOTH )
		szrDlgButtons.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )

		self.lblOutputSpace = wx.StaticText( self.pnlDlgButtons, wx.ID_ANY, u"Output Space:", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lblOutputSpace.Wrap( -1 )

		szrDlgButtons.Add( self.lblOutputSpace, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALIGN_RIGHT|wx.BOTTOM|wx.LEFT|wx.TOP, 0 )

		cboOutputSpaceChoices = [ u"PsychoPy RGB (rgb)" ]
		self.cboOutputSpace = wx.Choice( self.pnlDlgButtons, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, cboOutputSpaceChoices, 0 )
		self.cboOutputSpace.SetSelection( 0 )
		szrDlgButtons.Add( self.cboOutputSpace, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL|wx.EXPAND, 5 )


		szrDlgButtons.Add( ( 0, 0), 1, wx.EXPAND, 0 )

		self.cmdCancel = wx.Button( self.pnlDlgButtons, wx.ID_ANY, u"Cancel", wx.DefaultPosition, wx.DefaultSize, 0 )
		szrDlgButtons.Add( self.cmdCancel, 0, wx.BOTTOM|wx.LEFT|wx.TOP, 5 )

		self.cmdCopy = wx.Button( self.pnlDlgButtons, wx.ID_ANY, u"Copy", wx.DefaultPosition, wx.DefaultSize, 0 )
		szrDlgButtons.Add( self.cmdCopy, 0, wx.BOTTOM|wx.LEFT|wx.TOP, 5 )

		self.cmdInsert = wx.Button( self.pnlDlgButtons, wx.ID_ANY, u"Insert", wx.DefaultPosition, wx.DefaultSize, 0 )
		szrDlgButtons.Add( self.cmdInsert, 0, wx.ALL, 5 )


		self.pnlDlgButtons.SetSizer( szrDlgButtons )
		self.pnlDlgButtons.Layout()
		szrDlgButtons.Fit( self.pnlDlgButtons )
		szrMain.Add( self.pnlDlgButtons, 0, wx.ALL|wx.EXPAND, 5 )


		self.SetSizer( szrMain )
		self.Layout()

		self.Centre( wx.BOTH )

		# Connect Events
		self.Bind( wx.EVT_CLOSE, self.OnClose )
		self.sldRedChannel.Bind( wx.EVT_SCROLL, self.OnRedScroll )
		self.spnRedChannel.Bind( wx.EVT_SPINCTRLDOUBLE, self.OnRedSpin )
		self.spnRedChannel.Bind( wx.EVT_TEXT_ENTER, self.OnRedTextEnter )
		self.sldGreenChannel.Bind( wx.EVT_SCROLL, self.OnGreenScroll )
		self.spnGreenChannel.Bind( wx.EVT_SPINCTRLDOUBLE, self.OnGreenSpin )
		self.spnGreenChannel.Bind( wx.EVT_TEXT_ENTER, self.OnGreenTextEnter )
		self.sldBlueChannel.Bind( wx.EVT_SCROLL, self.OnBlueScroll )
		self.spnBlueChannel.Bind( wx.EVT_SPINCTRLDOUBLE, self.OnBlueSpin )
		self.spnBlueChannel.Bind( wx.EVT_TEXT_ENTER, self.OnBlueTextEnter )
		self.txtHexRGB.Bind( wx.EVT_KEY_DOWN, self.OnHexRGBKeyDown )
		self.rdoRGBModePsychoPy.Bind( wx.EVT_RADIOBUTTON, self.OnRGBModePsychoPy )
		self.rdoRGBModeNormalized.Bind( wx.EVT_RADIOBUTTON, self.OnRGBModeNormalized )
		self.rdoRGBMode255.Bind( wx.EVT_RADIOBUTTON, self.OnRGBMode255 )
		self.sldHueChannel.Bind( wx.EVT_SCROLL, self.OnHueScroll )
		self.spnHueChannel.Bind( wx.EVT_SPINCTRLDOUBLE, self.OnHueSpin )
		self.spnHueChannel.Bind( wx.EVT_TEXT_ENTER, self.OnHueTextEnter )
		self.sldStaturationChannel.Bind( wx.EVT_SCROLL, self.OnSaturationScroll )
		self.spnSaturationChannel.Bind( wx.EVT_SPINCTRLDOUBLE, self.OnSaturationSpin )
		self.spnSaturationChannel.Bind( wx.EVT_TEXT_ENTER, self.OnSaturationTextEnter )
		self.sldValueChannel.Bind( wx.EVT_SCROLL, self.OnValueScroll )
		self.spnValueChannel.Bind( wx.EVT_SPINCTRLDOUBLE, self.OnValueSpin )
		self.spnValueChannel.Bind( wx.EVT_TEXT_ENTER, self.OnValueTextEnter )
		self.lstColorPresets.Bind( wx.EVT_LISTBOX, self.OnPresetSelect )
		self.cmdCancel.Bind( wx.EVT_BUTTON, self.OnCancel )
		self.cmdCopy.Bind( wx.EVT_BUTTON, self.OnCopy )
		self.cmdInsert.Bind( wx.EVT_BUTTON, self.OnInsert )

	def __del__( self ):
		pass


	# Virtual event handlers, override them in your derived class
	def OnClose( self, event ):
		event.Skip()

	def OnRedScroll( self, event ):
		event.Skip()

	def OnRedSpin( self, event ):
		event.Skip()

	def OnRedTextEnter( self, event ):
		event.Skip()

	def OnGreenScroll( self, event ):
		event.Skip()

	def OnGreenSpin( self, event ):
		event.Skip()

	def OnGreenTextEnter( self, event ):
		event.Skip()

	def OnBlueScroll( self, event ):
		event.Skip()

	def OnBlueSpin( self, event ):
		event.Skip()

	def OnBlueTextEnter( self, event ):
		event.Skip()

	def OnHexRGBKeyDown( self, event ):
		event.Skip()

	def OnRGBModePsychoPy( self, event ):
		event.Skip()

	def OnRGBModeNormalized( self, event ):
		event.Skip()

	def OnRGBMode255( self, event ):
		event.Skip()

	def OnHueScroll( self, event ):
		event.Skip()

	def OnHueSpin( self, event ):
		event.Skip()

	def OnHueTextEnter( self, event ):
		event.Skip()

	def OnSaturationScroll( self, event ):
		event.Skip()

	def OnSaturationSpin( self, event ):
		event.Skip()

	def OnSaturationTextEnter( self, event ):
		event.Skip()

	def OnValueScroll( self, event ):
		event.Skip()

	def OnValueSpin( self, event ):
		event.Skip()

	def OnValueTextEnter( self, event ):
		event.Skip()

	def OnPresetSelect( self, event ):
		event.Skip()

	def OnCancel( self, event ):
		event.Skip()

	def OnCopy( self, event ):
		event.Skip()

	def OnInsert( self, event ):
		event.Skip()





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This demo was created by Andrew Schofield to show how to use the advanced input/output 
functionality in the bits.py module. 

It also acts as a test routine to make sure your device is working and that the 
bits.py module is intact. 

The bits.py module mostly provides wrapper functions for getting control of the
CRS hardware family, Bits++, Bits#, Display++ and their variants. ViSaGe devices
are not supported unless you have one with a Bits# emulation mode. 

Most but not all of the bits.py commands are tested. Similarly
most but not all of the CRS functionality is tested often in combination. But there 
are two many combinations to test every thing.

There is some support for driving stereo goggles (CRS FE1) via the D25 feature connector 
output. Bits# has a stereo goggles port which works differently and is not tested. Nor is 
support for it currently provided in bits.py.

Stereo via frame interleaving is not recommended on a Display++ due to LCD pixel transition speeds.
CRS have a 3D version of Display++ which uses polarised galsses and line interleaving. This is not
specifically implemented in bits.py yet but you can still build an stereo image yourself.

Bits++ support is relatively minimal with, in particular, digital inputs not being supported.

To work effectively on a Bits++ box you will need an Oscilloscope and:
    Connect DOUT1 (pin2) to Oscilloscope channel A"
    Connect DOUT5 (pin6) to Oscilloscope channel B"
You will then observe changes in those outputs.

To work effectively on all other devices you will need to:

    Connect DOUT1 (pin 2) to DIN7 (pin 21)
    Connect DOUT5 (pin 6) to DIN8 (pin 24)
    Connect DOUT6 (pin 7) to DIN9 (pin 25)

and connect Analog OUT 1 to Analog IN 1 if you have a Bits# or a Display++
with the optional Analog factures. This will cunningly allow the CRS device to monitor
its own outputs. Testing both output and input features at once.

You can select the CRS hardware and options in a dialog box.
Note Screen is the screen number of your CRS connected monitor = 0 or 1.
(your working in coder after all).

Enter None to see what happens if there is 
no device and hence no serial comms.

Note that the monitor and LUT settings don't really matter much.

You will get warnings about stuff being found on the input buffer. This is because the command to
stop the CRS device from sending data expects a return message to say that data collection has stopped
but will always find the last chunk of data recorded instead.

See xxxxx for a description of the bits.py approach to CRS hardware and programmer guide.

"""
from psychopy import locale_setup, sound, gui, visual, core, data, event, logging
from psychopy import monitors, filters, gamma 
from psychopy.hardware import crs
from scipy import misc

from psychopy.constants import (NOT_STARTED, STARTED, PLAYING, PAUSED,
                                STOPPED, FINISHED, PRESSED, RELEASED, FOREVER)
import numpy as np  # whole numpy lib is available, prepend 'np.'
from numpy import (sin, cos, tan, log, log10, pi, average,
                   sqrt, std, deg2rad, rad2deg, linspace, asarray)
from time import sleep, clock, time
import os  # handy system and path functions
import sys  # to get file system encoding


# Ensure that relative paths start from the same directory as this script
_thisDir = os.path.dirname(os.path.abspath(__file__))#.decode(sys.getfilesystemencoding())
os.chdir(_thisDir)

# Store info about the experiment session
expName = 'crsTest'  # from the Builder filename that created this script
expInfo = {'Device': 'Display++',
           'Analog': 'No',
           'Touch screen': 'Yes',
           'Button box': 'CB6',
           'Monitor': 'Display++160',
           'LUTfile': 'invGammaLUT.txt',
           'Screen': '1'}
dlg = gui.DlgFromDict(dictionary=expInfo, title=expName)
if dlg.OK == False:
    core.quit()  # user pressed cancel


#===================================================================#
# Setup the Window
print("Open window")
win = visual.Window(
    size=(800, 600), fullscr=True, screen=int(expInfo['Screen']),
    allowGUI=False, allowStencil=False,
    monitor=expInfo['Monitor'], color=[0,0,0], colorSpace='rgb',
    blendMode='avg', useFBO=True,
    units='deg')
# store frame rate of monitor if we can measure it
expInfo['frameRate'] = win.getActualFrameRate()
if expInfo['frameRate'] != None:
    frameDur = 1.0 / round(expInfo['frameRate'])
else:
    frameDur = 1.0 / 60.0  # could not measure, so guess

# Initialize components for Routine "preflight"
#preflightClock = core.Clock()


#frameRate=win.getActualFrameRate()
#print(expInfo['frameRate'])

mon=monitors.Monitor(expInfo['Monitor'],distance=56)

#=======================================================================================#
#Opening the appropriate a CRS class of the desired / necessary type
print("open CRS")
if expInfo['Device']=='Bits++':
    bits = crs.BitsPlusPlus(win, mode='bits++',rampType=1) 
if expInfo['Device']=='Bits#':
    bits = crs.BitsSharp(win, mode='bits++',checkConfigLevel=1) 
if expInfo['Device']=='Display++' or expInfo['Device']=='None':
    if expInfo['Touch screen']=="Yes":
        if expInfo['Device']=='Display++':
            bits = crs.DisplayPlusPlusTouch(win, mode='bits++',checkConfigLevel=1) 
        else:
            bits = crs.DisplayPlusPlusTouch(win, mode='bits++',checkConfigLevel=1,noComms=True) 
    else:
        if expInfo['Device']=='Display++':
            bits = crs.DisplayPlusPlus(win, mode='bits++',checkConfigLevel=1) 
        else:
            bits = crs.DisplayPlusPlus(win, mode='bits++',checkConfigLevel=1,noComms=True) 

#=======================================================================================#
# If Bits# or Display++ initialise the device.                                          #
# This can also be done via parameter setters but this illustrates the low level send   #
# commands.                                                                              #
if  expInfo['Device'] != 'Bits++':
    #bits = crs.BitsSharp(win, mode='bits++') 
    #gamma.setGamma(win.winHandle._dc, 1.0, 1)
    bits.sendMessage('$TemporalDithering=[ON]\r')
    bits.read(timeout=0.1)
    bits.sendMessage('$VideoFrameRate\r')
    bits.read(timeout=0.1)
    lutfile = expInfo['LUTfile']
    msg='$enableGammaCorrection=['+lutfile+']\r'
    bits.sendMessage(msg)
    bits.read(timeout=0.1)
    bits.sendMessage(r'$EnableTouchScreen=[OFF]\e')
    bits.read(timeout=0.1)
    bits.sendMessage('$Stop\r')
    # Clear the buffer
    bits.flush()
    bits.RTBoxDisable()
    bits.flush()
    print("OK I seem to have cleared the buffer now")

#=======================================================================================#
# House keeping
    print("Connect DOUT1 (pin 2) to DIN7 (pin 21)")
    print("Connect DOUT5 (pin 6) to DIN8 (pin 24)")
    print("Connect DOUT6 (pin 7) to DIN9 (pin 25)")
    if expInfo['Device'] == 'Bits#' or expInfo['Analog'] == 'Yes':
        print("Connect Analog OUT 1 TO Analog IN 1")
else:
    print("Connect DOUT1 (pin2) to Oscilloscope A")
    print("Connect DOUT5 (pin6) to Oscilloscope B")

sleep(10)

# Make a grating
grating = visual.GratingStim(
    win=win, name='grating',units='norm', 
    tex='sin', mask=None,
    ori=0, pos=(0,0), size=(1, 1), sf=5, phase=1.0,
    color=[1,1,1], colorSpace='rgb', opacity=1,blendmode='avg',
    texRes=128, interpolate=True, depth=-2.0)

bits.resetClock()
bits.win.flip()
sleep(1)

#=============================================================================#
#LUT test
# Varies the contrast of a grating by setting the software LUT in bits++ mode
# via a Tlock command.
print("1: LUT test")
print("Should see pulsating grating")
frameN=-1
while frameN < bits.frameRate*5:
    # get current time
    frameN = frameN + 1  # number of completed frames (so 0 is the first frame)
    # update/draw components on each frame
    #bits.setAnalog(AOUT1 = 5*sin(t*50), AOUT2 = 5*cos(t*50))
    # *thanksMsg* updates
    if frameN == 0:
        grating.setAutoDraw(True)
    bits.setContrast((sin(np.pi*frameN/bits.frameRate)))
    win.flip()
grating.setAutoDraw(False)
bits.setContrast(1.0)
win.flip()



#=============================================================================#
# Trigger and goggle tests
print("2a: Triggers test")
# Sets up a 2ms trigger pulse and issues via TLock 
# Recoding Digital inputs via the status logging commands
bits.setTrigger(0b0000000010,0.002,0.002)
bits.startTrigger()
bits.win.flip()
T=clock()
if expInfo['Device'] != 'Bits++': #Bist# / Display++ version
    # Recording Digital inputs via the status logging commands

    # Example of starting and stopping status log
    bits.startStatusLog()
    while clock()-T<5:
        bits.win.flip()
    bits.stopStatusLog()
    
    #Example of reading status values
    vals = bits.getAllStatusValues()
    if vals:
        if len(vals) > 20:
            print("All status values")
            for i in range(-20,0):
                print(vals[i])
            print("DIN7 should change lots")
    
    #Example of reading status events
    vals = bits.getAllStatusEvents()
    if vals:
        if len(vals) > 20:
            print("All status events")
            for i in range(-20,0):
                print(vals[i])
            print("Input 7 should have 2ms high every frame.")
else: # Bits++ version
    print("Should observe 2ms pulses on Ch A")
    while clock()-T<5:
        bits.win.flip()
bits.stopTrigger()
bits.win.flip()
sleep(5)
if expInfo['Device'] != 'Bits++':
    bits.flush()

#=============================================================================#
print("2b: Goggles test")
# Sets up alternating left / right goggle control via D25 output and the TLock
bits.startGoggles(left=0,right=1)
bits.win.flip()
T=clock()
if expInfo['Device'] != 'Bits++': # Bits# and Display++ version
    # Recording Digital inputs via the status logging commands
    
    # Example of starting and stopping status log
    bits.startStatusLog()
    while clock()-T<5:
        bits.win.flip()
    bits.stopStatusLog()
    bits.flush()
    
    #Example of reading status values
    vals = bits.getAllStatusValues()
    if vals:
        if len(vals) > 20:
            print("All status values")
            for i in range(-20,0):
                print(vals[i])
            print("DIN8 should change lots")
    
    #Example of reading status events
    vals = bits.getAllStatusEvents()
    if vals:
        if len(vals) > 20:
            print("All status events")
            for i in range(-20,0):
                print(vals[i])
            print("Input 8 should change state every frame.")
else: # Bits++ version
    print("Should observe square wave on Ch B")
    while clock()-T<5:
        bits.win.flip()

bits.stopGoggles()
bits.win.flip()
sleep(5)
if expInfo['Device'] != 'Bits++':
    bits.flush()

#=============================================================================#
print("2c: Goggles with  triggers")
# Drive the goggles and a trigger at the same time.
# Note uses trigger from 2a as this will have been kept
# even when the goggles were used ioin 2b.
bits.startTrigger()
bits.startGoggles()
bits.win.flip()
T=clock()
if expInfo['Device'] != 'Bits++':# Bits# and Display++ version
    # Recording Digital inputs via the status logging commands
    
    # Example of starting and stopping status log
    bits.startStatusLog()
    while clock()-T<5:
        bits.win.flip()
    bits.stopStatusLog()
    bits.flush()
    
    #Example of reading status values
    vals = bits.getAllStatusValues()
    if vals:
        if len(vals) > 20:
            print("All status values")
            for i in range(-20,0):
                print(vals[i])
            print("DIN7 and DIN8 should change lots")
            
    #Example of reading status events
    vals = bits.getAllStatusEvents()
    if vals:
        if len(vals) > 20:
            print("All status events")
            for i in range(-20,0):
                print(vals[i])
            print("Input 7 should have 2ms high every frame.")
            print("Input 8 should change state every frame.")
else: # Bits++ version
    print("Should observe 2ms pulses on Ch A")
    print("Should observe square wave on Ch B")
    while clock()-T<5:
        #sleep(0.002)
        bits.win.flip()

bits.stopGoggles()
bits.stopTrigger()
bits.win.flip()
sleep(5)

if expInfo['Device'] != 'Bits++':
    bits.flush()

#=============================================================================#
# If using a bits++ all available tests are now done
if expInfo['Device'] == 'Bits++':
    print("All tests done")
else: # otherwise carry on
    bits.flush() # Flush the com port often
    
    print("2d: Poll status")
    #=============================================================================#
    # Status polling
    # Example of using the pollStatus command
    bits.pollStatus()
    # Get the digital word form of DIN states
    valD = bits.getDigitalWord()
    # Example of how to print it
    if valD:
        print(bin(valD['DWORD']))
    
    #=============================================================================#
    #RTBox operation
    # Example of RTBox Enable, here using a bespoke mapping to route DIN9 to btn1
    print("2e: Single shot trigger detected by RTBox  via DIN")
    bits.RTBoxEnable(mode=['down'], map=[('btn1','Din9')])
    
    # Example of using send trigger to issue a trigger that will pulse Dout6 which would be connected to DIN
    bits.sendTrigger(0b1111111,0,0.004)
    bits.win.flip()  # win.flip needed to finish off the trigger

    #Example of using RTBoxKeysPressed to detect button box events.
    if not bits.noComms: # noComms safe
        # Wait for a key - its probably already been recvied anyway.
        while not bits.RTBoxKeysPressed(1):
            continue
    # Get the response
    btn=bits.getRTBoxResponse()
    # Example button response
    if btn:
        print(btn)
    else:
        print("No button")
    bits.flush()
    btn = None
    # Disable the RTBox for now
    bits.RTBoxDisable()
    sleep(5)

    #=============================================================================#
    # Example for making the CRS device beep
    print("3a: Beep test")
    bits.beep(400,0.5)
    sleep(0.5)
    
    #=============================================================================#
    # Using Tlock to reset the CRS device clock
    print("3b: Clock Reset test")
    bits.resetClock()
    bits.win.flip()
    sleep(1)
    T=clock()
    
    #=============================================================================#
    # Example for using PollStatus to read the device time
    bits.pollStatus()
    print("Time taken to poll status = ",clock()-T)
    val = bits.getStatus()
    if val:
        print("Time recorded by device should be about 1s. Actually = ", val.time)
    sleep(5)
    # Just to make sure the status log has been stopped
    bits.sendMessage('$Stop\r')
    bits.flush()

    #=============================================================================#
    # Example Using statusBox to record key presses
    
    print("4a: Using statusBox and Din")
    bits.statusBoxEnable(mode = ['up','IO10'])
    bits.sendTrigger(0b1111111,0,0.004)
    bits.win.flip()  # win.flip needed to finish off the trigger

    if not bits.noComms: # noComms safe
        # Wait for a key - its probably already been recvied anyway.
        while not bits.statusBoxKeysPressed(1):
            continue
    # Disable the status box as soon as you don't need it.
    bits.statusBoxDisable()
    bits.stopTrigger()
    bits.win.flip() # win.flip to make sure triggers are off.
    # Get the response
    btn = bits.getStatusBoxResponse() # just get first response and ditch others.
    # Example button response
    if btn:
        print(btn)
    else:
        print("No button")

    btn = None
    
    bits.flush()
    sleep(5)
    
    #=============================================================================#
    #   Example using the statusBox with different button boxes
    print("4b: Using statusBox with a button box")
    
    # Enables different statusBox defaults depending on users input
    if expInfo['Button box'] == 'CB6': #IR box
        bits.statusBoxEnable(mode=['CB6','Down'])
        print("Press a button on the Box")
        
        #Example statusBoxWait command - waits for a button press
        button = bits.statusBoxWait()
        if button:
            print(button)
    elif expInfo['Button box'] == 'IO6': # A wired box
        bits.statusBoxEnable(mode=['IO6','Down'])
        print("Press one of first 3 buttons on the Box")
        
        #Example statusBoxWait command - waits for a button press
        button = bits.statusBoxWait()
        if button:
            print(button)
    elif expInfo['Button box'] == 'IO': # A wired box with only 3 buttons
        bits.statusBoxEnable(mode=['IO','Down'])
        print("Press of first 3 buttons on the Box")
        
        #Example statusBoxWait command - waits for a button press
        button = bits.statusBoxWait()
        if button:
            print(button)
    
    #=============================================================================#
    #   Example using the statusBox to get multiple responses
    #   We've left the statusBix running for this
    print("4c: Multiple buttons")
    print("Now press some buttons")
    T = clock()
    while clock()-T<10:
        bits.win.flip()
    bits.statusBoxDisable()
    print('Some buttons presses ',bits.statusBoxKeysPressed())
    res = bits.getAllStatusBoxResponses()
    if res:
        print(len(res))
        print(res)
    bits.flush()

    #=============================================================================#
    # More RTBox usage examples
    print("4d: RTBox test")
    bits.flush()
    
    # Enables different RTBox defaults depending on users input
    if expInfo['Button box'] == 'CB6': #IR box
        bits.RTBoxEnable(mode=['CB6','Down'])
        print("Press a button on the Box")
        
        #Example RTBoxWait command - waits for a button press
        button = bits.RTBoxWait()
        if button:
            print(button)
    elif expInfo['Button box'] == 'IO6': # A wired box
        bits.RTBoxEnable(mode=['IO6','Down'])
        print("Press one of first 3 buttons on the Box")
        
        #Example RTBoxWait command - waits for a button press
        button = bits.RTBoxWait()
        if button:
            print(button)
    elif expInfo['Button box'] == 'IO': # A wired box
        bits.RTBoxEnable(mode=['IO','Down'])
        print("Press of first 3 buttons on the Box")
        
        #Example RTBoxWait command - waits for a button press
        button = bits.RTBoxWait()
        if button:
            print(button)
    
    #=============================================================================#
    # Example of how to calibrate or charaterise the RTBox timer
    # relative to the host clock
    print("4e: RTBox calibration test")
    res=bits.RTBoxCalibrate(5)
    bits.RTBoxDisable()
    bits.flush()
    print('Average clock difference =', res)
    
    #=============================================================================#
    # Commands for analog outputs, used if we have a Bits# or a 
    # display++ with analog feature
    if (expInfo['Device'] == 'Bits#' 
        or expInfo['Device'] == 'None' 
        or expInfo['Analog'] == 'Yes'):
        print("5: Analog tests")
        
        
        #=============================================================================#
        # Example of sending an analog out - will continue while we poll the status
        print("5a: Analog only using pollStatus")
        bits.sendAnalog(2,2)
        sleep(1) # let the output it settle
        bits.pollStatus()
        
        # Example for getting analog values and displaying them
        val = bits.getAnalog()
        bits.win.flip()
        #bits.stopAnalog
        print("Analog 1 should be = 2v")
        if val:
            print(val['ADC'])
        sleep(5)
        bits.flush()
        
        #=============================================================================#
        # Example using triggers and analog outputs at the same time.
        print("5b: Analog and triggers")
        # Example using set analog
        bits.setAnalog(3,3)
        # Set a long trigger
        bits.setTrigger(0b1111111101,0.0,0.0084)
        
        #Start both outputs
        bits.startTrigger()
        bits.startAnalog()
        bits.win.flip()
        sleep(1)
        bits.pollStatus()
        valA = bits.getAnalog()
        valD = bits.getDigitalWord()
        bits.stopAnalog
        bits.stopTrigger()
        print("Analog 1 should be = 3v")
        if valA:
            print(valA['ADC'])
        print("Digital word should have bit 8 low")
        if valD:
            print(bin(valD['DWORD']))
        sleep(5)
        bits.flush()
        
        #=============================================================================#
        # Example using triggers, goggles and analog all at once
        # Also test the ability of triggers to service people messing with
        # Goggles and Analog outs as these all use the same communication channel.
        print("5c: Goggles, analog and triggers")
        bits.setAnalog(3,3)
        bits.setTrigger(0b0000000010,0.002,0.002)
        
        # Test to see if triggers can service lots of toing a froing
        # of the analog and goggles outputs.
        bits.startAnalog()
        bits.win.flip()
        bits.startGoggles()
        bits.win.flip()
        bits.stopAnalog()
        bits.win.flip()
        bits.stopGoggles()
        bits.win.flip()
        
        # Now set them all going
        bits.startGoggles()
        bits.startAnalog()
        bits.startTrigger()
        bits.win.flip()
        T=time()
        
        #Setting the status event parameters to determine which events
        #are recorded. Now only 'up' events will be registered
        bits.setStatusEventParams(DINBase=0b1111111111, 
                                      IRBase=0b111111, 
                                      TrigInBase=0,
                                      ADCBase=0,
                                      threshold=1.0,
                                      mode=['up'])
        bits.startStatusLog()
        while time()-T<5:
            bits.setAnalog(5*sin((time()-T)), 0)
            bits.win.flip()
        bits.stopStatusLog()
        sleep(1)
        bits.flush()
        
        # Analog should show up on the status log
        vals = bits.getAllStatusValues()
        if vals:
            if len(vals) > 1:
                print("All status values")
                for i in range(1,len(vals),700):# only read every 700'th status report as 
                    # analog inputs update slowly relative to everything else
                    print(vals[i])
                print("Should see analog 1 changing")
                
        #The Goggles and triggers are best detected as events
        vals = bits.getAllStatusEvents()
        if vals:
            if len(vals) > 20:
                print("All status events")
                for i in range(-20,0):
                    print(vals[i])
            print("Input 7 should go up once every frame.")
            print("Input 8 should go up every other frame.")
        bits.stopTrigger()
        bits.stopGoggles()
        bits.stopAnalog()
        bits.win.flip()
        
        #=============================================================================#
        # Example for detecting changes on the analog inputs via status events
        # Not the analog threshold was set to 0.5 volts above
        print("5d: Detecting analog events via status")
        bits.setStatusEventParams(DINBase=0b1111111111, 
                                      IRBase=0b111111, 
                                      TrigInBase=0,
                                      ADCBase=0,
                                      threshold=1.0,
                                      mode=['Up'])
        bits.startStatusLog()
        bits.startAnalog()
        T = time()
        while time()-T<10:
            bits.setAnalog(5*sin((time()-T)), 0)
            bits.win.flip()
        bits.stopStatusLog()
        sleep(1)
        bits.stopAnalog()
        bits.flush()
        
        # Analog should show up on the status log
        vals = bits.getAllStatusValues()
        if vals:
            if len(vals) > 1:
                print("Status values")
                for i in range(1,len(vals),700):# only read every 700'th status report as 
                    # analog inputs update slowly relative to everything else
                    print(vals[i])
                print("Should see analog 1 changing")
                
        #Analog changes as events
        vals = bits.getAllStatusEvents()
        if vals:
            if len(vals) > 5:
                print("Status events")
                for i in range(len(vals)):
                    print(vals[i])
            print("Should see some events on Analog 1.")
            print("May get spurious events on some Bits# boxes.")
        bits.stopAnalog()
        bits.win.flip()
       
        #=============================================================================#
        # Example for usinf the statusBox to detect analog events
        print("5e: Using status box to detect analog events")
        bits.sendAnalog(0.5,0.5)
        bits.win.flip
        bits.win.flip
        bits.statusBoxEnable(mode=['Analog','Up','down'], threshold = 3.25)
        bits.sendAnalog(4.0,4.0)
        bits.win.flip
        sleep(2)
        # need to set analog low again by more than
        # threshold in order to record new event
        bits.sendAnalog(0.5,0.5)
        bits.win.flip
        sleep(2)
        bits.sendAnalog(4.0,4.0)
        bits.win.flip
        sleep(2)
        if not bits.noComms: # noComms safe
        # Wait for a key - its probably already been recvied anyway.
            while not bits.statusBoxKeysPressed():
                continue
        # Get the response
        bits.statusBoxDisable()
        btn = bits.getAllStatusBoxResponses()
        
        # Example button response
        if btn:
            print(btn)
        else:
            print("No button")
    
    #=============================================================================#
    # Touch screen tests
    if ((expInfo['Device'] == 'Display++' and expInfo['Touch screen'] == 'Yes')
          or expInfo['Device'] == 'None'):
        print("6: Touch Screen test")
        
        #=============================================================================#
        #Example of a touch enable, wait, disable cycle
        print("6a: Touch the screen")
        bits.touchEnable()
        val=bits.touchWait() # pause while waiting for touch
        bits.touchDisable()
        if val:
            print(val)
        sleep(5)
        
        #=============================================================================#
        # Example of a touch enable, touchPressed, disable cycle
        print("6b: Touch the screen again")
        bits.touchEnable()
        if not bits.noComms:
            while not bits.touchPressed(): # idle while waiting for touch
                # But you could do stuff here
                continue
        # Example of getAllTouchResponses()
        vals=bits.getAllTouchResponses()
        bits.touchDisable()
        if vals:
            for i in range(len(vals)):
                print(vals[i])
            
        sleep(5)
        
        #=============================================================================#
        # Exampe of using a touch log cycle to get lots of touch responses
        bits.startTouchLog()
        print("6c: Touch the screen lots")
        sleep(10)
        bits.stopTouchLog()
        
        #=============================================================================#
        # Example of getting the full touch log and then extracting events
        bits.getTouchLog()
        vals=bits.getTouchEvents()
        if vals:
            for i in range(len(vals)):
                print(vals[i])
    print("All tests done")
del bits
win.close()
core.quit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
import sys

import wx
import wx.html2

from psychopy.localization import _translate
from psychopy.projects import pavlovia
from psychopy import logging


class BaseFrame(wx.Frame):
    def __init__(self, *args, **kwargs):
        wx.Frame.__init__(self, *args, **kwargs)
        self.Center()
        # set up menu bar
        self.menuBar = wx.MenuBar()
        self.fileMenu = self.makeFileMenu()
        self.menuBar.Append(self.fileMenu, _translate('&File'))
        self.SetMenuBar(self.menuBar)

    def makeFileMenu(self):
        fileMenu = wx.Menu()
        app = wx.GetApp()
        keyCodes = app.keys
        # add items to file menu
        fileMenu.Append(wx.ID_CLOSE,
                        _translate("&Close View\t%s") % keyCodes['close'],
                        _translate("Close current window"))
        self.Bind(wx.EVT_MENU, self.closeFrame, id=wx.ID_CLOSE)
        # -------------quit
        fileMenu.AppendSeparator()
        fileMenu.Append(wx.ID_EXIT,
                        _translate("&Quit\t%s") % keyCodes['quit'],
                        _translate("Terminate the program"))
        self.Bind(wx.EVT_MENU, app.quit, id=wx.ID_EXIT)
        return fileMenu

    def closeFrame(self, event=None, checkSave=True):
        self.Destroy()

    def checkSave(self):
        """If the app asks whether everything is safely saved
        """
        return True  # for OK


class PavloviaMiniBrowser(wx.Dialog):
    """This class is used by to open an internal browser for the user stuff
    """
    style = wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER
    def __init__(self, parent, user=None, loginOnly=False, logoutOnly=False,
                 style=style, *args, **kwargs):
        # create the dialog
        wx.Dialog.__init__(self, parent, style=style, *args, **kwargs)
        # create browser window for authentication
        self.browser = wx.html2.WebView.New(self)
        self.loginOnly = loginOnly
        self.logoutOnly = logoutOnly
        self.tokenInfo = {}

        # do layout
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.browser, 1, wx.EXPAND, 10)
        self.SetSizer(sizer)
        if loginOnly:
            self.SetSize((600, 600))
        else:
            self.SetSize((700, 600))
        self.CenterOnParent()

        # check there is a user (or log them in)
        if not user:
            self.user = pavlovia.getCurrentSession().user
        if not user:
            self.login()
        if not user:
            return None

    def logout(self):
        self.browser.Bind(wx.html2.EVT_WEBVIEW_LOADED, self.checkForLogoutURL)
        self.browser.LoadURL('https://gitlab.pavlovia.org/users/sign_out')

    def login(self):
        self._loggingIn = True
        authURL, state = pavlovia.getAuthURL()
        self.browser.Bind(wx.html2.EVT_WEBVIEW_ERROR, self.onConnectionErr)
        self.browser.Bind(wx.html2.EVT_WEBVIEW_LOADED, self.checkForLoginURL)
        self.browser.LoadURL(authURL)

    def setURL(self, url):
        self.browser.LoadURL(url)

    def gotoUserPage(self):
        if self.user:
            url = self.user.attributes['web_url']
            self.browser.LoadURL(url)

    def editUserPage(self):
        url = "https://gitlab.pavlovia.org/profile"
        self.browser.LoadURL(url)
        self.SetSizeWH(1240, 840)

    def gotoProjects(self):
        self.browser.LoadURL("https://pavlovia.org/projects.html")

    def onConnectionErr(self, event):
        if 'INET_E_DOWNLOAD_FAILURE' in event.GetString():
            self.EndModal(wx.ID_EXIT)
            raise Exception("{}: No internet connection available.".format(event.GetString()))

    def checkForLoginURL(self, event):
        url = event.GetURL()
        if 'access_token=' in url:
            self.tokenInfo['token'] = self.getParamFromURL(
                'access_token', url)
            self.tokenInfo['tokenType'] = self.getParamFromURL(
                'token_type', url)
            self.tokenInfo['state'] = self.getParamFromURL(
                'state', url)
            self._loggingIn = False  # we got a log in
            self.browser.Unbind(wx.html2.EVT_WEBVIEW_LOADED)
            pavlovia.login(self.tokenInfo['token'])
            if self.loginOnly:
                self.EndModal(wx.ID_OK)
        elif url == 'https://gitlab.pavlovia.org/dashboard/projects':
            # this is what happens if the user registered instead of logging in
            # try now to do the log in (in the same session)
            self.login()
        else:
            logging.info("OAuthBrowser.onNewURL: {}".format(url))

    def checkForLogoutURL(self, event):
        url = event.GetURL()
        if url == 'https://gitlab.pavlovia.org/users/sign_in':
            if self.logoutOnly:
                self.EndModal(wx.ID_OK)

    def getParamFromURL(self, paramName, url=None):
        """Takes a url and returns the named param"""
        if url is None:
            url = self.browser.GetCurrentURL()
        return url.split(paramName + '=')[1].split('&')[0]


class PavloviaCommitDialog(wx.Dialog):
    """This class will be used to brings up a commit dialog
    (if there is anything to commit)"""

    def __init__(self, *args, **kwargs):

        # pop kwargs for Py2 compatibility
        changeInfo = kwargs.pop('changeInfo', '')
        initMsg = kwargs.pop('initMsg', '')

        super(PavloviaCommitDialog, self).__init__(*args, **kwargs)

        # Set Text widgets
        wx.Dialog(None, id=wx.ID_ANY, title=_translate("Committing changes"))
        self.updatesInfo = wx.StaticText(self, label=changeInfo)
        self.commitTitleLbl = wx.StaticText(self, label=_translate('Summary of changes'))
        self.commitTitleCtrl = wx.TextCtrl(self, size=(500, -1), value=initMsg)
        self.commitDescrLbl = wx.StaticText(self, label=_translate('Details of changes\n (optional)'))
        self.commitDescrCtrl = wx.TextCtrl(self, size=(500, 200), style=wx.TE_MULTILINE | wx.TE_AUTO_URL)

        # Set buttons
        self.btnOK = wx.Button(self, wx.ID_OK)
        self.btnCancel = wx.Button(self, wx.ID_CANCEL)

        # Format elements
        self.setToolTips()
        self.setDlgSizers()

    def setToolTips(self):
        """Set the tooltips for the dialog widgets"""
        self.commitTitleCtrl.SetToolTip(
            wx.ToolTip(
                _translate("Title summarizing the changes you're making in these files")))
        self.commitDescrCtrl.SetToolTip(
            wx.ToolTip(
                _translate("Optional details about the changes you're making in these files")))

    def setDlgSizers(self):
        """
        Set the commit dialog sizers and layout.
        """
        commitSizer = wx.FlexGridSizer(cols=2, rows=2, vgap=5, hgap=5)
        commitSizer.AddMany([(self.commitTitleLbl, 0, wx.ALIGN_RIGHT),
                             self.commitTitleCtrl,
                             (self.commitDescrLbl, 0, wx.ALIGN_RIGHT),
                             self.commitDescrCtrl])
        buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
        if sys.platform == "win32":
            btns = [self.btnOK, self.btnCancel]
        else:
            btns = [self.btnCancel, self.btnOK]
        buttonSizer.AddMany(btns)

        # main sizer and layout
        mainSizer = wx.BoxSizer(wx.VERTICAL)
        mainSizer.Add(self.updatesInfo, 0, wx.ALL | wx.EXPAND, border=5)
        mainSizer.Add(commitSizer, 1, wx.ALL | wx.EXPAND, border=5)
        mainSizer.Add(buttonSizer, 0, wx.ALL | wx.ALIGN_RIGHT, border=5)
        self.SetSizerAndFit(mainSizer)
        self.Layout()

    def ShowCommitDlg(self):
        """Show the commit application-modal dialog

        Returns
        -------
        wx event
        """
        return self.ShowModal()

    def getCommitMsg(self):
        """
        Gets the commit message for the git commit.

        Returns
        -------
        string:
            The commit message and description.
            If somehow the commit message is blank, a default is given.
        """
        if self.commitTitleCtrl.IsEmpty():
            commitMsg = "_"
        else:
            commitMsg = self.commitTitleCtrl.GetValue()
            if not self.commitDescrCtrl.IsEmpty():
                commitMsg += "\n\n" + self.commitDescrCtrl.GetValue()
        return commitMsg



from pathlib import Path
import subprocess
import re
import time, sys, os
import argparse
import shutil

thisFolder = Path(__file__).parent

with Path().home()/ 'keys/apple_ost_id' as p:
    IDENTITY = p.read_text().strip()
with Path().home()/ 'keys/apple_psychopy3_app_specific' as p:
    PWORD = p.read_text().strip()

ENTITLEMENTS = thisFolder / "entitlements.plist"
BUNDLE_ID = "org.opensciencetools.psychopy"
USERNAME = "admin@opensciencetools.org"

# handy resources for info:
#
# why use zip file to notarize as well as dmg:
#   https://deciphertools.com/blog/notarizing-dmg/
# notarize from Python:
#   https://github.com/najiji/notarizer/blob/master/notarize.py
# apple entitlements:
#     https://developer.apple.com/documentation/xcode/notarizing_macos_software_before_distribution/resolving_common_notarization_issues


class testDMG:
    def __init__(self, appFile, version, destination=None):
        self.appFile = Path(appFile)
        self.version = version
        self._dmgFile = None

    def dmgBuild(self):
        dmgFilename = str(self.appFile).replace(".app", "_rw.dmg")
        print(f"building dmg file: {dmgFilename}")
        # remove previous file if it's there
        if Path(dmgFilename).exists():
            os.remove(dmgFilename)
        # then build new one
        import dmgbuild
        icon = (thisFolder.parent /
                'psychopy/app/Resources/psychopy.icns').resolve()
        background = (thisFolder / "builtin-arrow.tiff").resolve()
        dmgbuild.build_dmg(
                filename=dmgFilename,
                volume_name=f'PsychoPy {self.version}',
                settings={
                    'format': 'UDRW',
                    'files': [str(thisFolder / 'dmg_settings.py')],
                    'symlinks': { 'Applications': '/Applications' },
                    'size': '3m', # but maybe irrelevant in UDRW mode?
                    'badge_icon': str(icon),
                    'background': None, #"builtin-arrow",
                    'icon_size': 128,
                    'icon_locations': {
                        'dmg_settings.py': (150, 160),
                        'Applications': (350, 160)
                    },
                    'window_rect': ((600, 600), (500, 400)),
                },
        )
        self._dmgFile = dmgFilename
        return dmgFilename

    def dmgCompress(self):
        dmgFinalFilename = str(self.appFile).replace(".app", f"_{self.version}.dmg")
        # remove previous file if it's there
        if Path(dmgFinalFilename).exists():
            os.remove(dmgFinalFilename)

        cmdStr = f"hdiutil convert {self._dmgFile} " \
                 f"-format UDZO " \
                 f"-o {dmgFinalFilename}"
        exitcode, output = subprocess.getstatusoutput(cmdStr)
        print(output)
        return dmgFinalFilename

def main():
        thistest = testDMG(appFile="psychoPy.app", version='20.2')
        thistest.dmgBuild()
        dmgFile = thistest.dmgCompress()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo of class data.ExperimentHandler
"""

from psychopy import data, logging
from numpy import random
logging.console.setLevel(logging.DEBUG)

exp = data.ExperimentHandler(name='testExp',
                version='0.1',
                extraInfo={'participant':'jwp', 'ori':45},
                runtimeInfo=None,
                originPath=None,
                savePickle=True,
                saveWideText=True,
                dataFileName='testExp')

# a first loop (like training?)
conds = data.createFactorialTrialList(
    {'faceExpression':['happy', 'sad'], 'presTime':[0.2, 0.3]})
training = data.TrialHandler(trialList=conds, nReps=3, name='train',
                 method='random',
                 seed=100)  # this will set the global seed - for the whole exp
exp.addLoop(training)
# run those trials
for trial in training:
    training.addData('training.rt', random.random() * 0.5 + 0.5)
    if random.random() > 0.5:
        training.addData('training.key', 'left')
    else:
        training.addData('training.key', 'right')
    exp.nextEntry()

# then run 3 repeats of a staircase
outerLoop = data.TrialHandler(trialList=[], nReps=3, name='stairBlock',
                 method='random')
exp.addLoop(outerLoop)
for thisRep in outerLoop:  # the outer loop doesn't save any data
    staircase = data.StairHandler(startVal=10, name='staircase', nTrials=5)
    exp.addLoop(staircase)
    for thisTrial in staircase:
        id=random.random()
        if random.random() > 0.5:
            staircase.addData(1)
        else:
            staircase.addData(0)
        exp.addData('id', id)
        exp.nextEntry()
for e in exp.entries:
    print(e)
print("Done. 'exp' experimentHandler will now (end of script) save data to testExp.csv")
print(" and also to testExp.psydat, which is a pickled version of `exp`")

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Functions and classes related to unit conversion respective to a particular
monitor"""

from psychopy import monitors
from psychopy import logging
import numpy as np
import re
from numpy import array, tan, pi, radians, hypot, degrees, arctan
from math import hypot as hypot3d

# Maps supported coordinate unit type names to the function that converts
# the given unit type to PsychoPy OpenGL pix unit space.
_unit2PixMappings = dict()

# the following are to be used by convertToPix


def _pix2pix(vertices, pos, win=None):
    return pos + vertices
_unit2PixMappings['pix'] = _pix2pix
_unit2PixMappings['pixels'] = _pix2pix


def _cm2pix(vertices, pos, win):
    return cm2pix(pos + vertices, win.monitor)
_unit2PixMappings['cm'] = _cm2pix


def _deg2pix(vertices, pos, win):
    return deg2pix(pos + vertices, win.monitor)
_unit2PixMappings['deg'] = _deg2pix
_unit2PixMappings['degs'] = _deg2pix


def _degFlatPos2pix(vertices, pos, win):
    posCorrected = deg2pix(pos, win.monitor, correctFlat=True)
    vertices = deg2pix(vertices, win.monitor, correctFlat=False)
    return posCorrected + vertices
_unit2PixMappings['degFlatPos'] = _degFlatPos2pix


def _degFlat2pix(vertices, pos, win):
    return deg2pix(array(pos) + array(vertices), win.monitor,
                   correctFlat=True)
_unit2PixMappings['degFlat'] = _degFlat2pix


def _norm2pix(vertices, pos, win):
    if win.useRetina:
        return (pos + vertices) * win.size / 4.0
    else:
        return (pos + vertices) * win.size / 2.0

_unit2PixMappings['norm'] = _norm2pix


def _height2pix(vertices, pos, win):
    if win.useRetina:
        return (pos + vertices) * win.size[1] / 2.0
    else:
        return (pos + vertices) * win.size[1]

_unit2PixMappings['height'] = _height2pix


def posToPix(stim):
    """Returns the stim's position in pixels,
    based on its pos, units, and win.
    """
    return convertToPix([0, 0], stim.pos, stim.win.units, stim.win)


def convertToPix(vertices, pos, units, win):
    """Takes vertices and position, combines and converts to pixels
    from any unit

    The reason that `pos` and `vertices` are provided separately is that
    it allows the conversion from deg to apply flat-screen correction to
    each separately.

    The reason that these use function args rather than relying on
    self.pos is that some stimuli use other terms (e.g. ElementArrayStim
    uses fieldPos).
    """
    unit2pixFunc = _unit2PixMappings.get(units)
    if unit2pixFunc:
        return unit2pixFunc(vertices, pos, win)
    else:
        msg = "The unit type [{0}] is not registered with PsychoPy"
        raise ValueError(msg.format(units))


def addUnitTypeConversion(unitLabel, mappingFunc):
    """Add support for converting units specified by unit_label to pixels
    to be used by convertToPix (therefore a valid unit for your PsychoPy
    stimuli)

    mapping_func must have the function prototype:

    def mapping_func(vertices, pos, win):
        # Convert the input vertices, pos to pixel positions PsychoPy
        # will use for OpenGL call.

        # unit type -> pixel mapping logic here
        # .....

        return pix
    """
    if unitLabel in _unit2PixMappings:
        msg = "The unit type label [{0}] is already registered with PsychoPy"
        raise ValueError(msg.format(unitLabel))
    _unit2PixMappings[unitLabel] = mappingFunc


# Built in conversion functions follow ...


def cm2deg(cm, monitor, correctFlat=False):
    """Convert size in cm to size in degrees for a given Monitor object
    """
    # check we have a monitor
    if not isinstance(monitor, monitors.Monitor):
        msg = ("cm2deg requires a monitors.Monitor object as the second "
               "argument but received %s")
        raise ValueError(msg % str(type(monitor)))
    # get monitor dimensions
    dist = monitor.getDistance()
    # check they all exist
    if dist is None:
        msg = "Monitor %s has no known distance (SEE MONITOR CENTER)"
        raise ValueError(msg % monitor.name)
    if correctFlat:
        return np.degrees(np.arctan(cm / dist))
    else:
        return cm / (dist * 0.017455)


def deg2cm(degrees, monitor, correctFlat=False):
    """Convert size in degrees to size in pixels for a given Monitor object.

    If `correctFlat == False` then the screen will be treated as if all
    points are equal distance from the eye. This means that each "degree"
    will be the same size irrespective of its position.

    If `correctFlat == True` then the `degrees` argument must be an Nx2 matrix
    for X and Y values (the two cannot be calculated separately in this case).

    With `correctFlat == True` the positions may look strange because more
    eccentric vertices will be spaced further apart.
    """
    # check we have a monitor
    if not hasattr(monitor, 'getDistance'):
        msg = ("deg2cm requires a monitors.Monitor object as the second "
               "argument but received %s")
        raise ValueError(msg % str(type(monitor)))
    # get monitor dimensions
    dist = monitor.getDistance()
    # check they all exist
    if dist is None:
        msg = "Monitor %s has no known distance (SEE MONITOR CENTER)"
        raise ValueError(msg % monitor.name)
    if correctFlat:
        rads = radians(degrees)
        cmXY = np.zeros(rads.shape, 'd')  # must be a double (not float)
        if rads.shape == (2,):
            x, y = rads
            cmXY[0] = hypot(dist, tan(y) * dist) * tan(x)
            cmXY[1] = hypot(dist, tan(x) * dist) * tan(y)
        elif len(rads.shape) > 1 and rads.shape[1] == 2:
            cmXY[:, 0] = hypot(dist, tan(rads[:, 1]) * dist) * tan(rads[:, 0])
            cmXY[:, 1] = hypot(dist, tan(rads[:, 0]) * dist) * tan(rads[:, 1])
        else:
            msg = ("If using deg2cm with correctedFlat==True then degrees "
                   "arg must have shape [N,2], not %s")
            raise ValueError(msg % (repr(rads.shape)))
        # derivation:
        #    if hypotY is line from eyeball to [x,0] given by
        #       hypot(dist, tan(degX))
        #    then cmY is distance from [x,0] to [x,y] given by
        #       hypotY * tan(degY)
        #    similar for hypotX to get cmX
        # alternative:
        #    we could do this by converting to polar coords, converting
        #    deg2cm and then going back to cartesian,
        #    but this would be slower(?)
        return cmXY
    else:
        # the size of 1 deg at screen centre
        return np.array(degrees) * dist * 0.017455


def cm2pix(cm, monitor):
    """Convert size in cm to size in pixels for a given Monitor object.
    """
    # check we have a monitor
    if not isinstance(monitor, monitors.Monitor):
        msg = ("cm2pix requires a monitors.Monitor object as the"
               " second argument but received %s")
        raise ValueError(msg % str(type(monitor)))
    # get monitor params and raise error if necess
    scrWidthCm = monitor.getWidth()
    scrSizePix = monitor.getSizePix()
    if scrSizePix is None:
        msg = "Monitor %s has no known size in pixels (SEE MONITOR CENTER)"
        raise ValueError(msg % monitor.name)
    if scrWidthCm is None:
        msg = "Monitor %s has no known width in cm (SEE MONITOR CENTER)"
        raise ValueError(msg % monitor.name)

    return cm * scrSizePix[0] / float(scrWidthCm)


def pix2cm(pixels, monitor):
    """Convert size in pixels to size in cm for a given Monitor object
    """
    # check we have a monitor
    if not isinstance(monitor, monitors.Monitor):
        msg = ("cm2pix requires a monitors.Monitor object as the second"
               " argument but received %s")
        raise ValueError(msg % str(type(monitor)))
    # get monitor params and raise error if necess
    scrWidthCm = monitor.getWidth()
    scrSizePix = monitor.getSizePix()
    if scrSizePix is None:
        msg = "Monitor %s has no known size in pixels (SEE MONITOR CENTER)"
        raise ValueError(msg % monitor.name)
    if scrWidthCm is None:
        msg = "Monitor %s has no known width in cm (SEE MONITOR CENTER)"
        raise ValueError(msg % monitor.name)
    return pixels * float(scrWidthCm) / scrSizePix[0]


def deg2pix(degrees, monitor, correctFlat=False):
    """Convert size in degrees to size in pixels for a given Monitor object
    """
    # get monitor params and raise error if necess
    scrWidthCm = monitor.getWidth()
    scrSizePix = monitor.getSizePix()
    if scrSizePix is None:
        msg = "Monitor %s has no known size in pixels (SEE MONITOR CENTER)"
        raise ValueError(msg % monitor.name)
    if scrWidthCm is None:
        msg = "Monitor %s has no known width in cm (SEE MONITOR CENTER)"
        raise ValueError(msg % monitor.name)

    cmSize = deg2cm(degrees, monitor, correctFlat)
    return cmSize * scrSizePix[0] / float(scrWidthCm)


def pix2deg(pixels, monitor, correctFlat=False):
    """Convert size in pixels to size in degrees for a given Monitor object
    """
    # get monitor params and raise error if necess
    scrWidthCm = monitor.getWidth()
    scrSizePix = monitor.getSizePix()
    if scrSizePix is None:
        msg = "Monitor %s has no known size in pixels (SEE MONITOR CENTER)"
        raise ValueError(msg % monitor.name)
    if scrWidthCm is None:
        msg = "Monitor %s has no known width in cm (SEE MONITOR CENTER)"
        raise ValueError(msg % monitor.name)
    cmSize = pixels * float(scrWidthCm) / scrSizePix[0]
    return cm2deg(cmSize, monitor, correctFlat)



# Minimal Oculus Rift head-mounted display example. Press the 'q' key or use
# the application GUI to exit. Requires PsychXR to be installed.
#
# This file is public domain.
#
from psychopy import visual, event, core  # visual imported, even if not used!

# Create a VR session, treat the returned object just like a regular window.
#
hmd = visual.Rift()

# loop until the user quits the app through the GUI menu
stopApp = False
while not stopApp:
    # You must call these functions before drawing anything. First, you need to
    # get the current tracking state of the head, then pass the pose to
    # `calcEyePoses` to compute the eye positions. Only then you can begin
    # rendering scenes. If `calcEyePoses` is not called, the application will
    # stall.
    trackingState = hmd.getTrackingState()
    hmd.calcEyePoses(trackingState.headPose.thePose)

    for i in ('left', 'right'):
        hmd.setBuffer(i)  # select the eye buffer to draw to

        # Setup the viewing parameters for the current buffer, this needs to be
        # called every time the buffer changes.
        #
        # For standard PsychoPy stimuli (e.g. GratingStim, ImageStim, etc.) you
        # should use 'setDefaultView' with 'mono=True' when creating a
        # visual.Rift instance. This configures the headset to properly render
        # 2D stimuli, treating the HMD as a monitor.
        #
        hmd.setDefaultView()

    # send the rendered buffer to the HMD
    hmd.flip()

    # check if the application should exit
    if event.getKeys('q') or hmd.shouldQuit:
        stopApp = True

# cleanly end the session
core.quit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Example of reading events from an iohub hdf5 file, saving to a tab delimited text file.

Events are split into trials by reading the time of 'TRIAL_START' and
'TRIAL_END' experiment Message events.

SAVE_EVENT_TYPE and SAVE_EVENT_FIELDS specify the event type, and which event fields, are saved.

This example can process hdf5 files saved by running the gcCursor demo.
"""
from psychopy.iohub.datastore.util import saveEventReport

# Specify which event type and event fields to save
SAVE_EVENT_TYPE = 'MonocularEyeSampleEvent'
SAVE_EVENT_FIELDS = ['time', 'gaze_x', 'gaze_y', 'pupil_measure1', 'status']
#SAVE_EVENT_TYPE = 'BinocularEyeSampleEvent'
#SAVE_EVENT_FIELDS = ['time', 'left_gaze_x', 'left_gaze_y', 'left_pupil_measure1',
#                     'right_gaze_x', 'right_gaze_y', 'right_pupil_measure1', 'status']

# Specify the experiment message text used to split events into trial periods.
# Set both to None to save all events.
TRIAL_START_MESSAGE = "TRIAL_START"
TRIAL_END_MESSAGE = "TRIAL_END"

if __name__ == '__main__':
    result = saveEventReport(hdf5FilePath=None, eventType=SAVE_EVENT_TYPE, eventFields=SAVE_EVENT_FIELDS,
                             trialStart=TRIAL_START_MESSAGE, trialStop=TRIAL_END_MESSAGE)
    if result:
        file_saved, events_saved = result
        print("Saved %d events to %s." % (events_saved, file_saved))
    else:
        raise RuntimeError("saveEventReport failed.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Test for using iohub mouse with multiple windows on multiple monitors.

To enable multi window support for the iohub Mouse device:

launchHubServer(window=win, Mouse=dict(enable_multi_window=True))

In this mode, if the mouse is over a psychopy window, mouse position is returned
as the pix position within the window, with origin (0,0) at window center. mouse event
window_id will equal a psychopy window handle (pyglet only).

Note: If testing using 'cm' or 'deg' units, use your actual monitor configurations
or update the `monitors.Monitor` created below to match your setup.

If the mouse is not over a psychopy window, desktop mouse position is returned with
window_id = 0.
"""
import sys

from psychopy import visual, core, monitors
from psychopy.iohub import launchHubServer
from psychopy.iohub.constants import EventConstants

# True = print mouse events to stdout, False = do not
PRINT_MOUSE_EVENTS = False

# Test creating a monitor before starting iohub
mon0 = monitors.Monitor('monitor0')
mon0.setDistance(60.0)
mon0.setWidth(33.0)
mon0.setSizePix((1280, 1024))
win = visual.Window((400, 400), pos=(0, 30), units='norm', fullscr=False, allowGUI=True, screen=0, monitor=mon0,
                    winType='pyglet')

# start the iohub server
io = launchHubServer(window=win, Mouse=dict(enable_multi_window=True))

# Test creating a monitor after starting iohub
mon1 = monitors.Monitor('monitor1')
mon1.setDistance(60.0)
mon1.setWidth(34.5)
mon1.setSizePix((1920, 1080))
win2 = visual.Window((600, 600), pos=(500, 30), units='cm', fullscr=False, allowGUI=True, screen=1, monitor=mon1,
                     winType='pyglet')

# access the iohub keyboard and mouse
keyboard = io.devices.keyboard
mouse = io.devices.mouse

# TODO: How to handle setPos when mouse enable_multi_window=True
# mouse.setPosition((0.0, 0.0))

txt_proto = 'Desktop x,y: {}, {}\nWin x,y: {:.4}, {:.4}\n\nwin.units: {}\n\n\nPress Any Key to Quit.'
win_stim={}
for w in visual.window.openWindows:
    win_stim[w()._hw_handle] = visual.TextStim(w(), pos=(-0.75, 0.0), units='norm', alignText='left', anchorHoriz='left',
                                               anchorVert='center', height=.1, autoLog=False, wrapWidth=1.5,
                                               text=txt_proto.format('?', '?', '?', '?', w().units))

io.clearEvents('all')

demo_timeout_start = core.getTime()

# Run the example until a keyboard event is received.
kb_events = None
while not kb_events:
    for stim in win_stim.values():
        stim.draw()
    win.flip()  # redraw the buffer
    flip_time = win2.flip()  # redraw the buffer

    # Check for Mouse events
    mouse_events = mouse.getEvents()
    if mouse_events:
        # Simple example of handling different mouse event types.
        if PRINT_MOUSE_EVENTS:
            for me in mouse_events:
                if me.type == EventConstants.MOUSE_MOVE:
                    print(me)
                elif me.type == EventConstants.MOUSE_DRAG:
                    print(me)
                elif me.type == EventConstants.MOUSE_BUTTON_PRESS:
                    print(me)
                elif me.type == EventConstants.MOUSE_BUTTON_RELEASE:
                    print(me)
                elif me.type == EventConstants.MOUSE_SCROLL:
                    print(me)
                else:
                    print("Unhandled event type:", me.type, me)
                    sys.exit()

        # Only update display based on last received event
        me = mouse_events[-1]
        #print("display: ", me.display_id)
        if me.window_id > 0:
            for win_handle, stim in win_stim.items():
                if win_handle != me.window_id:
                    stim.text = txt_proto.format('?', '?', '?', '?', stim.win.units)
                else:
                    stim.text = txt_proto.format('?', '?', me.x_position, me.y_position, stim.win.units)

        else:
            for win_handle, stim in win_stim.items():
                stim.text = txt_proto.format(me.x_position, me.y_position, '?', '?', stim.win.units)

        demo_timeout_start = mouse_events[-1].time
    # If 15 seconds passes without receiving any kb or mouse event,
    # then exit the demo
    if flip_time - demo_timeout_start > 15.0:
        print("Ending Demo Due to 15 Seconds of Inactivity.")
        break

    # Check for keyboard events.
    kb_events = keyboard.getEvents()

win.close()
win2.close()

core.quit()
# End of Example

# The contents of this file are in the public domain.



from pathlib import Path

import pytest
from psychopy import visual
from .test_basevisual import _TestColorMixin, _TestUnitsMixin
from .. import utils


class TestCircle(_TestColorMixin, _TestUnitsMixin):

    @classmethod
    def setup_class(self):

        self.win = visual.Window([128,128], pos=[50,50], allowGUI=False, autoLog=False)
        self.obj = visual.Circle(self.win, units="pix", pos=(0, 0), size=(128, 128), lineWidth=10)

        # Pixel which is the border color
        self.borderPoint = (64, 0)
        self.borderUsed = True
        # Pixel which is the fill color
        self.fillPoint = (64, 64)
        self.fillUsed = True
        # Shape has no foreground color
        self.foreUsed = False

    def test_radius(self):
        # Define some use cases
        cases = [
            {'size': [.1, .1], 'radius': 1, 'units': 'height',
             'minvertices': [-.1, -.1], 'maxvertices': [.1, .1],
             'label': ".1height"},
            {'size': [.1, .1], 'radius': 2, 'units': 'height',
             'minvertices': [-.2, -.2], 'maxvertices': [.2, .2],
             'label': ".2height"},
            {'size': [.2, .2], 'radius': 0.5, 'units': 'height',
             'minvertices': [-.1, -.1], 'maxvertices': [.1, .1],
             'label': ".1height"},

            {'size': [10, 10], 'radius': 1, 'units': 'pix',
             'minvertices': [-10, -10], 'maxvertices': [10, 10],
             'label': "10pix"},
            {'size': [10, 10], 'radius': 2, 'units': 'pix',
             'minvertices': [-20, -20], 'maxvertices': [20, 20],
             'label': "20pix"},
            {'size': [20, 20], 'radius': 0.5, 'units': 'pix',
             'minvertices': [-10, -10], 'maxvertices': [10, 10],
             'label': "10pix"},
        ]
        # Test each case
        for case in cases:
            self.win.flip()
            # Apply units, size and radius
            self.obj.units = case['units']
            self.obj.radius = case['radius']
            self.obj.size = case['size']
            # Check that this results in correct vertices
            verts = getattr(self.obj._vertices, case['units'])
            for i in [0, 1]:
                assert min(verts[:, i]) == case['minvertices'][i]
                assert max(verts[:, i]) == case['maxvertices'][i]
            # Check that vertices are drawn correctly
            filename = f"test_circle_radius_{case['label']}.png"
            self.obj.draw()
            # self.win.getMovieFrame(buffer='back').save(Path(utils.TESTS_DATA_PATH) / filename)
            # utils.compareScreenshot(Path(utils.TESTS_DATA_PATH) / filename, self.win, crit=20)
